/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998-1999  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    AVFS module (vfs layer)
*/

#include "op_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>


/*
   -- Meaning of error values --

   EFAULT: Internal error
   ENODEV: Function not implemented on this VFS device
*/


#define AVFS_MAJOR 238 /* FIXME: This is just a random value */

extern char **environ;

ave __av_dummyv;

struct avfs_private __av_priv = {0, 0, 0};

#define AV_MAXMAJOR 18
#define AV_MAXMINOR 256


static int loglevel = -1;
static char logname[AVFS_MAX_PATHLEN];

/* Since major is a function on SVR4, we can't use `ifndef major'.  */
#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#define HAVE_MAJOR
#endif
#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#define HAVE_MAJOR
#endif
#ifdef major			/* Might be defined in sys/types.h.  */
#define HAVE_MAJOR
#endif

#ifdef HAVE_VSNPRINTF
#define VSNPRINTF vsnprintf
#else 
#  ifdef HAVE___VSNPRINTF
#  define VSNPRINTF __vsnprintf
#  endif
#endif



#ifndef HAVE_MAJOR
#define major(dev)  (((dev) >> 8) & 0xff)
#define minor(dev)  ((dev) & 0xff)
#define makedev(maj, min)  (((maj) << 8) | (min))
#endif
#undef HAVE_MAJOR


#ifdef WITH_XMEM

#define MALLOCID 43
extern void *_iamalloc(size_t size, int id);
extern void *_iarealloc(void *ip, size_t size, int id);
extern void _iafree(void *ip, int id);
extern void amalloc_show_unfreed(int id);

#else

#ifdef HAVE___LIBC_MALLOC
#define MPREF(x) __libc_ ## x
extern void *__libc_malloc(size_t size);
extern void *__libc_realloc(void *ip, size_t size);
extern void __libc_free(void *ip);
#else
#define MPREF(x) x
#endif

#define _iamalloc(nbyte, id)      MPREF(malloc) (nbyte)
#define _iarealloc(ptr, nyte, id) MPREF(realloc) (ptr, nbyte)
#define _iafree(ptr, id)          MPREF(free) (ptr)
#define amalloc_show_unfreed(id) 

#endif


int __av_add_vdev(ave *v, struct vdev_info *newdev)
{
  if(__av_priv.vdev_num >= AV_MAXMAJOR) return -1;

  if(__av_priv.vdev_num >= __av_priv.vdevs_size) {
    struct vdev_info **newvdevs;
    int newsize;

    newsize = __av_priv.vdevs_size + 16;
    newvdevs = (struct vdev_info **)
      __av_realloc(v , __av_priv.vdevs, sizeof(*__av_priv.vdevs) * newsize);

    if(newvdevs == NULL) return -1;
    __av_priv.vdevs = newvdevs;
    __av_priv.vdevs_size = newsize;
  }
  newdev->minors = (struct pred_dev *)
    __av_malloc(v, sizeof(struct pred_dev) * AV_MAXMINOR);
  if(newdev->minors == NULL) return -1;
  newdev->minors[0].next = 0;
  newdev->minors[0].prev = 0;

  __av_priv.vdevs[__av_priv.vdev_num++] = newdev;
  return __av_priv.vdev_num - 1;
}


int __av_getminor(int major, avdev_t odev, avino_t oino)
{
  int i, n;
  int found;
  struct pred_dev *minors;

  minors = __av_priv.vdevs[major]->minors;

  found = 0;
  for(n = 1, i = minors[0].next; i; n++, i = minors[i].next) {
    if(minors[i].dev == odev && minors[i].ino == oino) {
      found = 1;
      break;
    }
  }

  if(!found && n == AV_MAXMINOR) i = minors[0].prev;

  if(found || n == AV_MAXMINOR) {
    minors[minors[i].prev].next = minors[i].next;
    minors[minors[i].next].prev = minors[i].prev;
  }
  else i = n;

  if(!found) {
    minors[i].dev = odev;
    minors[i].ino = oino;
    minors[i].minor = i | (__av_random() << 8);
  }

  minors[i].next = minors[0].next;
  minors[i].prev = 0;
  minors[minors[0].next].prev = i;
  minors[0].next = i;

  return minors[i].minor;
}

avdev_t __av_vmakedev(int major, int minor)
{
  return __av_mkdev(AVFS_MAJOR + major, minor & 0xFF);
  /* FIXME: Something better */
}


#if 0
int avfs_setflags(int add, int del)
{
  avfs_flags |= add;
  avfs_flags &= ~del;

  return avfs_flags;
}
#endif


void *__av_malloc(ave *v, avsize_t nbyte)
{
  char *p;

  p = (char *) _iamalloc(nbyte, MALLOCID);
  if(p == NULL) v->errn = ENOMEM;
  
  return (void *) p;
}

void *__av_realloc(ave *v, void *ptr, avsize_t nbyte)
{
  char *p;

  p = (char *) _iarealloc(ptr, nbyte, MALLOCID);
  if(p == NULL) v->errn = ENOMEM;
  
  return (void *) p;
}


void __av_free(void *ptr)
{
  if(ptr != NULL)
    _iafree(ptr, MALLOCID);
}

void __av_init_misc(void)
{
/*   FD_ZERO(&openfiles); */
/*   amalloc_show_unfreed(MALLOCID); */
}

int __av_pipe(ave *v, int pip[2])
{
  int res;

  res = pipe(pip);
  if(res == -1) v->errn = EIO;
  return res;
}

int __av_set_nonblock(ave *v, int fd)
{
  return fcntl(fd, F_SETFL, O_NONBLOCK);
}

int __av_localselect(ave *v, int n, void *readfds, void *writefds,
		     void *exceptfds, struct avtimeval *timeout)
{
  struct timeval tv;
  int res;
  
  if(timeout != AVNULL) {
    tv.tv_sec = timeout->sec;
    tv.tv_usec = timeout->usec;
  }

  errno = v->errn;
  res  = select(n, (fd_set *) readfds, (fd_set *) writefds,
		(fd_set *) exceptfds, timeout ? &tv : NULL);
  v->errn = errno;

  return res;
}

char *__av_strdup(ave *v, const char *s)
{
  char *ns;
  
  ns = (char *) __av_malloc(v, strlen(s) + 1);
  if(ns != NULL) strcpy(ns, s);

  return ns;
}


int __av_strncasecmp(const char *s1, const char *s2, avsize_t n)
{
#ifdef HAVE_STRNCASECMP
  return strncasecmp(s1, s2, n);
#else
  for(; *s1 && *s2 && n; s1++, s2++, n++) if(*s1 != *s2) break;
  return (*s1 - *s2);
#endif
}

avdev_t __av_mkdev(int major, int minor)
{
  return makedev(major, minor);
}

void __av_copy_filename(ave *v, char *dst, const char *src)
{
  strncpy(dst, src, AVFS_MAX_NAMELEN);
  dst[AVFS_MAX_NAMELEN] = '\0';
}

int __av_random(void)
{
  return (int) rand();
}

static void init_logfile(void)
{
  char *logstr;
  int lognameset = 0;

  logstr = getenv("AVFSDEBUG");
  if(logstr != NULL && isdigit((int) logstr[0]) && isdigit((int) logstr[1])) {
    loglevel = (logstr[0] - '0') * 10 + (logstr[1] - '0');
    if(logstr[2] && strlen(logstr) < AVFS_MAX_PATHLEN) {
      strcpy(logname, logstr + 2);
      lognameset = 1;
    }
  }

  if(loglevel < 0) 
    loglevel = AVLOG_WARNING;
  
  if(!lognameset) 
    strcpy(logname, "/dev/console");

}

#define LOGMSG_SIZE 1024
void __av_log(int level, const char *format, ...)
{
  va_list ap;
  int logfd;
  char buf[LOGMSG_SIZE];
  int off;
  int l;

  if(loglevel < 0) init_logfile();
  if(level > loglevel) return;

  va_start(ap, format);

  sprintf(buf, "AVFS {%u}:                ", (unsigned) getpid());
  l = strlen(buf);
  off = AV_MAX(15, l-15);

#ifdef VSNPRINTF
  VSNPRINTF(buf+off, LOGMSG_SIZE-30, format, ap);
#else
  /* Note: this is UNSECURE, and can casue BUFFER OVERFLOWS */
  vsprintf(buf+off, format, ap);
#endif
  
  va_end(ap);

  l = strlen(buf);
  buf[l] = '\n';
  buf[l+1] = '\0';

  logfd = __av_localopen(DUMMYV, logname, 
			 AVO_WRONLY | AVO_CREAT | AVO_APPEND , 0644);
  __av_localwrite(DUMMYV, logfd, buf, strlen(buf));
  __av_localclose(DUMMYV, logfd);
}

int __av_get_logfile(ave *v)
{
  const char *reallogname;
  int logfd;

  if(loglevel < 0) init_logfile();

  if(loglevel <= AVLOG_WARNING) reallogname = "/dev/null";
  else reallogname = logname;

  logfd = __av_localopen(DUMMYV, reallogname, 
			 AVO_WRONLY | AVO_CREAT | AVO_APPEND, 0644);

  return logfd;
}

void __av_default_stat(struct avstat *stbuf)
{
  stbuf->dev = 0;
  stbuf->ino = 0;
  stbuf->mode = 0;
  stbuf->nlink = 0;
  stbuf->uid = 0;
  stbuf->gid = 0;
  stbuf->rdev = 0;
  stbuf->size = 0;
  stbuf->blksize = 512;
  stbuf->blocks = 0;
  stbuf->atime = 0;
  stbuf->mtime = 0;
  stbuf->ctime = 0;
}

avtime_t __av_gettime(void)
{
  return time(NULL);
}

unsigned int __av_sleep(unsigned int seconds)
{
  return sleep(seconds);
}

avtime_t __av_mktime(struct avtm *tp)
{
  struct tm tms;
  
  tms.tm_sec  = tp->sec;
  tms.tm_min  = tp->min;
  tms.tm_hour = tp->hour;
  tms.tm_mday = tp->day;
  tms.tm_mon  = tp->mon;
  tms.tm_year = tp->year;
  tms.tm_isdst = -1;

  return mktime(&tms);
}

int __av_isroot(ave *v, vpath *vp)
{
  const char *params;
  struct stat stbuf1, stbuf2;

  if(vp != AVNULL && BASE(vp) == AVNULL && (params = EXTRA(vp)) != AVNULL) {
    if(params[0] == DIR_SEP_CHAR && !params[1]) return 0;

    /* FIXME: this is a kludge, the parser should do this */
    if(!params[0]) params = ".";

    if(__av_realstat(params, &stbuf1) != -1 && 
       __av_realstat("/", &stbuf2) != -1 && 
       stbuf1.st_dev == stbuf2.st_dev && 
       stbuf1.st_ino == stbuf2.st_ino) return 0;
  }

  v->errn = ENOENT;
  return -1;
}


struct vdev_info *__av_new_vdev(ave *v)
{
  struct vdev_info *vdev;
  AV_NEW(v, vdev);
  if(vdev == AVNULL) return AVNULL;

  vdev->devdata    = AVNULL;
  vdev->name       = AVNULL;
  vdev->exts       = AVNULL;
  vdev->flags      = 0;

  vdev->minors     = AVNULL;
  vdev->destroy    = AVNULL;

  vdev->resolve    = AVNULL;

  vdev->open       = AVNULL;
  vdev->close      = AVNULL;
  vdev->read       = AVNULL;
  vdev->write      = AVNULL;
  vdev->lseek      = AVNULL;
  vdev->fstat      = AVNULL;
  vdev->readdir    = AVNULL;

  vdev->stat       = AVNULL;
  vdev->access     = AVNULL;
  vdev->readlink   = AVNULL;
  
  vdev->unlink     = AVNULL;
  vdev->rmdir      = AVNULL;
  vdev->mkdir      = AVNULL;
  vdev->mknod      = AVNULL;
  vdev->rename     = AVNULL;
  vdev->link       = AVNULL;
  vdev->symlink    = AVNULL;

  vdev->chmod      = AVNULL;
  vdev->chown      = AVNULL;
  vdev->utime      = AVNULL;
  vdev->fchmod     = AVNULL;
  vdev->fchown     = AVNULL;

  vdev->fgetfile   = AVNULL;

  return vdev;
}

int __av_unlink_recursive(const char *file)
{
  int res;
  DIR *dirp;
  struct dirent *ent;
  struct stat stbuf;
  char *name;

  res = __av_localunlink(DUMMYV, file);
  if(res == 0) return 0;

  res = __av_realrmdir(file);
  if(res == 0) return 0;

  res = __av_reallstat(file,&stbuf);
  if(res == -1) return -1;
  
  if(!S_ISDIR(stbuf.st_mode)) return -1;

  dirp = opendir(file);
  if(dirp == NULL) return -1;

  while((ent = readdir(dirp)) != NULL) {
    name = ent->d_name;
    
    if(name[0] != '.' || (name[1] && (name[1] != '.' || name[2]))) {
      char *newname;

      newname = __av_malloc(DUMMYV, strlen(file) + 1 + strlen(name) + 1);
      if(newname == NULL) return -1;
      sprintf(newname, "%s/%s", file, name);

      __av_unlink_recursive(newname);
      __av_free(newname);
    }
  }
  closedir(dirp);

  return __av_realrmdir(file);
}

void __av_del_tmpfile(char *tmpf)
{
  if(tmpf != NULL) {
    AVDEB((AVLOG_DEBUG, "Deleting Tmp file: %s (usage: %i)", 
	   tmpf, __av_priv.tmpdirusage - 1));

    if(__av_realunlink(tmpf) == -1) __av_realrmdir(tmpf);
    __av_free(tmpf);
    __av_priv.tmpdirusage --;
  }

  /* FIXME: Should be under lock */

  if(__av_priv.tmpdir != NULL && !__av_priv.tmpdirusage && 
     __av_realrmdir(__av_priv.tmpdir) == 0) {
    AVDEB((AVLOG_DEBUG, "Deleting Tmp directory: %s", __av_priv.tmpdir));
    __av_free(__av_priv.tmpdir);
    __av_priv.tmpdir = NULL;
  }
}

char *__av_get_tmpfile(ave *v)
{
  char *tmpf;
  static int ctr = 0;
  
  if(__av_priv.tmpdir == AVNULL) {
    char *tmpdir;

    tmpdir = __av_malloc(v, 64);
    if(tmpdir == AVNULL) return AVNULL;

    strcpy(tmpdir, "/tmp/.avfs_tmp_XXXXXX");
    mktemp(tmpdir);
    if(tmpdir[0] == '\0') {
      __av_free(tmpdir);
      v->errn = ENFILE;
      return AVNULL;
    }
    
    AVDEB((AVLOG_DEBUG, "Creating Tmp directory: %s", tmpdir));

    if(__av_realmkdir(tmpdir, 0700) == -1) {
      __av_free(tmpdir);
      v->errn = ENFILE;
      return AVNULL;
    }
    __av_priv.tmpdirusage = 0;
    __av_priv.tmpdir = tmpdir;
  }

  tmpf = __av_malloc(v, 64);
  if(tmpf == AVNULL) return AVNULL;

  sprintf(tmpf, "%s/atmp%06i", __av_priv.tmpdir, ctr++);

  AVDEB((AVLOG_DEBUG, "Creating Tmp file: %s", tmpf));

  __av_priv.tmpdirusage ++;
  return tmpf;
}

int __av_get_tmpfd(ave *v)
{
  char *tmpf;
  int tmpfd;

  tmpf = __av_get_tmpfile(v);
  if(tmpf == AVNULL) return -1;
  
  tmpfd = __av_localopen(v, tmpf, AVO_RDWR | AVO_CREAT | AVO_EXCL, 0600);
  __av_del_tmpfile(tmpf);
  
  return tmpfd;
}

avoff_t __av_generic_lseek(ave *v, avoff_t *ptrp, avoff_t size,
			   avoff_t offset, int whence)
{
  avoff_t newptr;
  
  switch(whence) {
  case AVSEEK_SET:
    newptr = offset;
    break;
    
  case AVSEEK_CUR:
    newptr = *ptrp + offset;
    break;
    
  case AVSEEK_END:
    newptr = size + offset;
    break;

  default:
    v->errn = EINVAL;
    return -1;
  }
  
  if(newptr < 0) {
    v->errn = EINVAL;
    return -1;
  }
  
  *ptrp = newptr;
  
  return newptr;
}

void __av_registerfd(int fd)
{
  if(FD_OK(fd)) __av_priv.file_table[fd].type |= FINTERNAL;
  fcntl(fd, F_SETFD, FD_CLOEXEC);
}

void __av_closeall(void)
{
#if 0
  int i;

  /* Warning, this function is perfectly useless, but it has been the
     source of numerous bugs!!! */
  
  /* Latest bug was that sometimes some of (0,1,2) was virtual, and so
     it was closed after dup2. This way the child process got invalid
     fd as stdin/out/err */

  for(i = 3; i < AVFS_MAX_FILES; i++) 
    if(__av_priv.file_table[i].type & (FINTERNAL | FVIRTUAL)) 
      __av_realclose(i);

#endif
}

void __av_init_proginfo(struct proginfo *pi)
{
  pi->prog = AVNULL;
  pi->pid = -1;

  pi->wd = AVNULL;
  pi->ifd = -1;
  pi->ofd = -1;
  pi->efd = -1;
}

#ifdef AVFS_DEBUG
static char *catparams(const char *const *strarr)
{
  static char params[1024];
  int len, plen;
  
  len = 0;
  params[0] = 0;
  while(*strarr != AVNULL) {
    plen = strlen(*strarr);
    if(len + plen + 1 > 1023) break;
    strcpy(params + len, *strarr);
    len += plen;
    strcpy(params + len, " ");
    len ++;
    strarr++;
  }
  
  return params;
}
#endif

int __av_start_prog(ave *v, struct proginfo *pi)
{
  AVDEB((AVLOG_DEBUG, "Starting program %s", catparams(pi->prog)));

  pi->pid = fork();
  
  if(pi->pid == -1) {
    v->errn = EIO;
    return -1;
  }

  if(pi->pid == 0) { /* Child */
    char **envp = environ;
    int i;

    for(i = 0; envp[i] != NULL; i++) {
      if(strncmp(envp[i], "PWD=", 4) == 0)
	envp[i][4] = '\0';
      else if(strncmp(envp[i], "AVFSDEBUG=", 10) == 0)
	envp[i][10] = '\0';
    }

    set_vcwd("");

    if(pi->wd != AVNULL) 
      __av_realchdir(pi->wd);
    
    setsid(); /* Don't want to kill my parent if something goes wrong */

    /* FIXME: This is where this process gets Killed in Solaris tcsh 6.04.00, 
       during filename completion. Actually closing any one of 0, 1, 2
       has the same effect. WHY????????? */

    if(pi->ifd != -1) dup2(pi->ifd, 0);
    if(pi->ofd != -1) dup2(pi->ofd, 1);
    if(pi->efd != -1) dup2(pi->efd, 2);
    
    __av_closeall();

    execvp(pi->prog[0], (char **) pi->prog);
    fprintf(stderr, "Failed to exec %s\n", pi->prog[0]);
    exit(1);
  }
  return 0;
}

int __av_wait_prog(ave *v, struct proginfo *pi, int tokill, int check)
{
  int retv, wres;

  if(pi->pid != -1) {
    if(check) {
      wres = waitpid(pi->pid, &retv, WNOHANG);
      if(wres == 0) return 0;
    }
    else {
      if(tokill) kill(pi->pid, SIGKILL);
      wres = waitpid(pi->pid, &retv, 0);
    }

    /* FIXME: We only get the return status if the Program hasn't 
       waited on our child (like emacs does) */
    if(wres != -1) {
      if(!WIFEXITED(retv) || WEXITSTATUS(retv) != 0) {
	v->errn = EIO;
	return -1;
      }
    }
  }
  pi->pid = -1;

  return 1;
}
