/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    AVFSSTAT module

    This module exports some useful information about the AVFS state. 
*/

#include "config.h"

#include "vdev.h"
#include "avfsint.h"
#include "version.h"

#include <string.h>
#include <stdio.h>
#include <assert.h>

#define F2INO(f) ((f) + 10)

#ifndef LINUX_GLIBC
#define LINUX_GLIBC 0
#endif

typedef struct {
  avdev_t device;
} avfsstat_devd;

typedef struct {
  int function;
  avfsstat_devd *dd;
  char *messagebuf;
  avoff_t size;
  avoff_t ptr;
} avfsstat_fdi;

enum {
  FUNC_MAIN,
  FUNC_LIBVER,
  FUNC_HAND,
  FUNC_VCWD,
};

static int decode_function(const char *params)
{
  if(!params[0]) return FUNC_MAIN;
  if(params[0] != '-') return -1;
  if(!params[1]) return -1;
  if(params[2]) return -1;
  
  switch(params[1]) {
  case 'v':
    return FUNC_LIBVER;

  case 'm':
    return FUNC_HAND;

  case 'p':
    return FUNC_VCWD;
    
  }
  return -1;
}

static int get_function(ave *v, vpath *path)
{
  int function;
  const char *params;


  if(__av_isroot(v, BASE(path)) == -1) return -1;
  
  params = EXTRA(path);
  if(params[0] == DIR_SEP_CHAR) {
    v->errn = ENOTDIR;
    return -1;
  }
  
  function = decode_function(params);
  
  if(function == -1) {
    v->errn = ENOENT;
    return -1;
  }

  return function;
}

static char *list_handlers(ave *v)
{
  int vi, ei;
  struct ext_info *exts;
  char *msg;
  char *s;
  int msgsize;

  msgsize = 0;
  for(vi = 0; vi < __av_priv.vdev_num; vi++) {
    msgsize += strlen(__av_priv.vdevs[vi]->name) + 2;
    exts = __av_priv.vdevs[vi]->exts;
    if(exts != AVNULL) 
      for(ei = 0; exts[ei].from != AVNULL; ei++) {
	msgsize += strlen(exts[ei].from) + 1;
	if(exts[ei].to != AVNULL) 
	  msgsize += strlen(exts[ei].to) + 2;
      }
    msgsize += 1;
  }
  
  msg = __av_malloc(v, msgsize + 1);
  if(msg == AVNULL) return AVNULL;
  msg[0] = '\0';
  
  s = msg;
  for(vi = 0; vi < __av_priv.vdev_num; vi++) {
    sprintf(s, "%s:\t", __av_priv.vdevs[vi]->name);
    s += strlen(s);
    exts = __av_priv.vdevs[vi]->exts;
    if(exts != AVNULL) 
      for(ei = 0; exts[ei].from != AVNULL; ei++) {
	strcpy(s, exts[ei].from);
	s += strlen(s);
	if(exts[ei].to != AVNULL) {
	  sprintf(s, "(%s)", exts[ei].to);
	  s += strlen(s);
	}
	strcpy(s, " ");
	s += strlen(s);
      }
    strcpy(s, "\n");
    s += strlen(s);
  }

  /*   assert(strlen(msg) == msgsize); */
  return msg;
}


static char *init_message(ave *v, int function)
{
  char buf[1024];
  char *msg = AVNULL;

  switch(function) {
  case FUNC_MAIN:
    sprintf(buf, 
	    "AVFS Virtual File System (C) Miklos Szeredi, Version %s,\n"
	    "This library comes with ABSOLUTELY NO WARRANTY.\n", 
	    AVFS_VERSION);
    break;
    
  case FUNC_LIBVER:
    sprintf(buf, 
	    "Version: %s\n"
	    "Library Version: %i.%i\n"
	    "Compile date: %s\n"
	    "Compile system: %s%s\n",
	    AVFS_VERSION, VER_MAJOR, VER_MINOR, 
	    COMPILE_DATE, 
	    COMPILE_SYSTEM, LINUX_GLIBC ? " GLIBC" : "");
    break;

  case FUNC_HAND:
    msg = list_handlers(v);
    if(msg == AVNULL) return AVNULL;
    break;
    
  case FUNC_VCWD:
    msg = __av_malloc(v, strlen(__av_priv.vcwd) + 2);
    if(msg == AVNULL) return AVNULL;
    sprintf(msg, "%s\n", __av_priv.vcwd);
    break;

  default:
    v->errn = EFAULT;
    return AVNULL;
  }

  if(msg == AVNULL) {
    msg = __av_malloc(v, strlen(buf) + 1);
    if(msg == AVNULL) return AVNULL;
    strcpy(msg, buf);
  }

  return msg;
}


static int avfsstat_close(ave *v, void *devinfo)
{
  avfsstat_fdi *di = (avfsstat_fdi *) devinfo;

  __av_free(di->messagebuf);
  __av_free(di);

  return 0;
}

static void *avfsstat_open(ave *v, void *devdata, vpath *path, 
		       int flags, int mode)
{
  avfsstat_fdi *di = AVNULL;
  avfsstat_devd *dd = (avfsstat_devd *) devdata;
  int function;

  function = get_function(v, path);
  if(function == -1) return AVNULL;

  if((flags & AVO_ACCMODE) != AVO_RDONLY) {
    v->errn = EACCES;
    return AVNULL;
  }

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;
  
  di->function = function;
  di->dd = dd;
  di->ptr = 0;
  
  di->messagebuf = init_message(v, function);
  if(di->messagebuf == AVNULL) {
    __av_free(di);
    return AVNULL;
  }

  di->size = strlen(di->messagebuf);
  
  return (void *) di;
}

static avssize_t avfsstat_read(ave *v, void *devinfo, char *buf,
			       avsize_t nbyte)
{
  avfsstat_fdi *di = (avfsstat_fdi *) devinfo;
  avoff_t nact;
  
  if(di->ptr >= di->size) return 0;

  nact = AV_MIN(nbyte, (avsize_t) (di->size - di->ptr));
  
  memcpy(buf, di->messagebuf + di->ptr, nact);
  di->ptr += nact;
  
  return nact;
}


static avssize_t avfsstat_write(ave *v, void *devinfo, const char *buf,
				avsize_t nbyte)
{
  v->errn = EBADF;
  return -1;
}

static avoff_t avfsstat_lseek(ave *v, void *devinfo, avoff_t offset,
			      int whence)
{
  avfsstat_fdi *di = (avfsstat_fdi *) devinfo;

  return __av_generic_lseek(v, &di->ptr, di->size, offset, whence);
}


static void set_stat(struct avstat *buf)
{
  __av_default_stat(buf);

  buf->mode    = 0444 | AV_IFREG;
  buf->nlink   = 1;
  buf->uid     = 0;
  buf->gid     = 0;
  buf->size    = 0;
  buf->blocks  = 0;
  buf->atime   = __av_gettime();
  buf->mtime   = buf->atime;
  buf->ctime   = buf->atime;
}


static int avfsstat_fstat(ave *v, void *devinfo, struct avstat *buf, int flags)
{
  avfsstat_fdi *di = (avfsstat_fdi *) devinfo;

  set_stat(buf);
  buf->size = di->size;
  buf->dev = di->dd->device;
  buf->ino = F2INO(di->function);

  return 0;
}

static int avfsstat_access(ave *v, void *devdata, vpath *path, int amode)
{
  if(get_function(v, path) == -1) return -1;

  if(amode & (AVW_OK | AVX_OK)) {
    v->errn = EACCES;
    return -1;
  }

  return 0;
}


static void avfsstat_destroy(struct vdev_info *vdev)
{
  avfsstat_devd *dd = (avfsstat_devd *) vdev->devdata;

  __av_free(dd);
  __av_free(vdev->name);
  __av_free(vdev);
}


extern int __av_init_module_avfsstat(ave *v);

int __av_init_module_avfsstat(ave *v)
{
  avfsstat_devd *dd;
  struct vdev_info *avfsstat_vdev;
  int major;
  
  avfsstat_vdev = __av_new_vdev(v);
  if(avfsstat_vdev == AVNULL) return -1;
  
  avfsstat_vdev->open     = avfsstat_open;
  avfsstat_vdev->close    = avfsstat_close;
  avfsstat_vdev->read     = avfsstat_read;
  avfsstat_vdev->write    = avfsstat_write;
  avfsstat_vdev->lseek    = avfsstat_lseek;
  avfsstat_vdev->fstat    = avfsstat_fstat;
  avfsstat_vdev->access   = avfsstat_access;

  avfsstat_vdev->destroy  = avfsstat_destroy;

  if((avfsstat_vdev->name = __av_strdup(v, "avfsstat")) == AVNULL) goto error;
  
  avfsstat_vdev->flags = 0;

  AV_NEW(v, dd);
  if(dd == AVNULL) goto error;

  avfsstat_vdev->devdata = (void *) dd;

  major = __av_add_vdev(v, avfsstat_vdev);
  if(major == -1) goto error;
  
  dd->device = __av_vmakedev(major, 0);
  return major;

 error:
  avfsstat_destroy(avfsstat_vdev);
  return -1;
}
