/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)
    
    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    FILTPROG

    This module starts, stops filter programs (like gzip), and it reads
    data from the standard output and writes data to the standard input of
    this program.
*/

#include "filtprog.h"

#define F_BUFSIZ 16384
#define R_BUFSIZ 16384

#include <sys/time.h>
#include <string.h>

typedef struct {
  struct filt_params fp;
  void *devinfo;

  int opip, ipip;
  avoff_t rptr;
  avoff_t size;
  
  struct proginfo pri;

  int started;
  int wbufat;
  int wbuflen;
  char wbuf[F_BUFSIZ];

  fd_set readfds;
  fd_set writefds;

  int cachefd;
  avoff_t cacheptr;
} filtdi;

static void stop_filter(filtdi *di)
{
  if(!di->started) return;
  di->started = 0;

  if(di->ipip != -1) __av_localclose(DUMMYV, di->ipip);
  if(di->opip != -1) __av_localclose(DUMMYV, di->opip);

  di->ipip = di->opip = -1;

  __av_wait_prog(DUMMYV, &di->pri, 1, 0);
}

static int start_filter(ave *v, filtdi *di)
{
  int pipout[2];
  int pipin[2];
  int res;
  int logfd;

  di->started = 1;
  logfd = __av_get_logfile(v);

  if(__av_pipe(v, pipout) == -1) {
    stop_filter(di);
    return -1;
  }

  di->opip = pipout[0];

  if(__av_pipe(v, pipin) == -1) {
    __av_localclose(DUMMYV, pipout[1]);
    stop_filter(di);
    return -1;
  }

  di->ipip = pipin[1];

  __av_registerfd(di->ipip);
  __av_registerfd(di->opip);

  __av_set_nonblock(DUMMYV, di->opip);
  __av_set_nonblock(DUMMYV, di->ipip);

  di->pri.prog = di->fp.prog;
  di->pri.ifd  = pipin[0];
  di->pri.ofd  = pipout[1];
  di->pri.efd  = logfd;
  
  res = __av_start_prog(v, &di->pri);
  
  __av_localclose(DUMMYV, pipout[1]);
  __av_localclose(DUMMYV, pipin[0]);
  
  if(res == -1) {
    stop_filter(di);
    return -1;
  }

  __av_localclose(DUMMYV, logfd);
  
  di->rptr = 0;  
  di->wbuflen = 0;
  di->wbufat = 0;

  if((*di->fp.lseekfunc)(v, di->devinfo, 0, AVSEEK_SET) == -1) {
    stop_filter(di);
    return -1;
  }
  
  return 0;
}


static avssize_t read_from_pipe(ave *v, filtdi *di, char *buf, avsize_t nbyte)
{
  int numread;
  int res;

  for(numread = 0; nbyte > 0;) {
    if(di->ipip != -1 && di->wbufat == di->wbuflen) {
      res = (*di->fp.readfunc)(v, di->devinfo, di->wbuf, F_BUFSIZ);
      if(res == -1) return -1;
      if(res == 0) {
        __av_localclose(DUMMYV, di->ipip);
        di->ipip = -1;
      }
      else {
        di->wbuflen = res;
        di->wbufat = 0;
      }
    }
      
    FD_ZERO(&di->readfds);
    FD_ZERO(&di->writefds);
    FD_SET(di->opip, &di->readfds);
    if(di->ipip != -1) FD_SET(di->ipip, &di->writefds);

    res = __av_localselect(DUMMYV, AV_MAX(di->ipip, di->opip)+1, 
			   (void *) &di->readfds, (void *) &di->writefds, 
			   AVNULL, AVNULL);
    if(res == -1) {
      if(errno == EINTR) continue;
      v->errn = EIO;
      return -1;
    }
    
    if(FD_ISSET(di->opip, &di->readfds)) {
      avoff_t num;
      char readbuf[R_BUFSIZ];
      char *rbuf;
      
      if(buf == AVNULL) {
	num = AV_MIN(nbyte, R_BUFSIZ);
	rbuf = readbuf;
      }
      else {
	num = nbyte;
	rbuf = buf;
      }
      
      res = __av_localread(v, di->opip, rbuf, num);
      
      if(res == -1) {
        if(v->errn == EINTR || v->errn == EAGAIN) continue;
        return -1;
      }
      
      if(res == 0) {
        if(numread) break;

	if(__av_wait_prog(v, &di->pri, 0, 0) == -1) return -1;

	/* FIXME: If the file is all cached, close the read end */

	if(__av_wait_prog(v, &di->pri, 0, 0) == -1) return -1;	
	di->size = di->rptr;
        return 0;
      }

      if(di->cachefd != -1) {
	if(di->cacheptr != di->rptr) 
	  di->cacheptr = __av_locallseek(DUMMYV, di->cachefd, di->rptr, 
					 AVSEEK_SET);
	if(di->cacheptr != di->rptr || 
	   __av_localwrite(DUMMYV, di->cachefd, rbuf, res) != res) {

	  __av_log(AVLOG_WARNING, "Filtprog: Error writing to cachefile");
	  __av_localclose(DUMMYV, di->cachefd);
	  di->cachefd = -1;
	}
	else di->cacheptr += res;
      }

      if(buf != AVNULL) buf += res;
      numread += res;
      nbyte -= res;
      di->rptr += res;
    }
    
    if(di->ipip != -1 && FD_ISSET(di->ipip, &di->writefds)) {

      /* FIXME: do something with SIGPIPE */
      res = __av_localwrite(v, di->ipip, di->wbuf + di->wbufat, 
			    di->wbuflen - di->wbufat);
      if(res == -1) {
        if(v->errn == EINTR || v->errn == EAGAIN) continue;
        return -1;
      }
      
      di->wbufat += res;
    }
  }

  return numread;
}

avoff_t __av_filt_getsize(ave *v, void *devinfo)
{
  filtdi *di = (filtdi *) devinfo; 
  avssize_t res;
  int ires;
  
  if(di->size != -1) return di->size;

  if(!di->started || di->pri.pid == -1) {
    stop_filter(di);
    ires = start_filter(v, di);
    if(ires == -1) return -1;
  }
  
  do {
    res = read_from_pipe(v, di, NULL, R_BUFSIZ);
    if(res == -1) return -1;
  } while(res > 0);

  return di->size;
}


int __av_filt_getcachefile(ave *v, void *devinfo, avoff_t till)
{
  filtdi *di = (filtdi *) devinfo;
  avssize_t res;
  
  if(di->cachefd == -1) {
    v->errn = ENODEV;
    return -1;
  }

  if(di->rptr < till) {
    res = read_from_pipe(v, di, AVNULL, till - di->rptr);
    if(res == -1) return -1;
  }
  
  return di->cachefd;
}


avssize_t __av_filt_read(ave *v, void *devinfo, char *buf, avoff_t offset,
			 avsize_t nbyte)
{
  filtdi *di = (filtdi *) devinfo;
  avssize_t rres, numread;
  int ires;

  numread = 0;

  if(di->started && offset < di->rptr && di->cachefd != -1) {
    avoff_t nact, avail;

    avail = di->rptr - offset;
    nact = AV_MIN((avoff_t) nbyte, avail);

    if(di->cacheptr != offset) 
      di->cacheptr = __av_locallseek(DUMMYV, di->cachefd, offset, 
				     AVSEEK_SET);
    
    if(di->cacheptr != offset || 
       __av_localread(DUMMYV, di->cachefd, buf, nact) != nact) {

      __av_log(AVLOG_WARNING, "Filtprog: Error reading from cachefile");
      __av_localclose(DUMMYV, di->cachefd);
      di->cachefd = -1;
    }
    else {
      di->cacheptr += nact;
      numread += nact;
      offset += nact;
      buf += nact;
      nbyte -= nact;
    }
  }

  if(nbyte == 0) return numread;

  if(!di->started || di->pri.pid == -1 || offset < di->rptr) {
    stop_filter(di);
    ires = start_filter(v, di);
    if(ires == -1) return -1;
  }

  if(offset != di->rptr) {
    rres = read_from_pipe(v, di, AVNULL, offset - di->rptr);
    if(rres == -1) return -1;
    
    if(offset != di->rptr) return 0;
  }

  rres = read_from_pipe(v, di, buf, nbyte);
  if(rres == -1) return -1;
  
  numread += rres;
  
  return numread;
}

void *__av_open_filtprog(ave *v, struct filt_params *fp, void *devinfo)
{
  filtdi *di;

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;

  di->fp = *fp;
  di->devinfo = devinfo;

  __av_init_proginfo(&(di->pri));

  di->ipip = -1;
  di->opip = -1;
  di->started = 0;
  
  di->size = -1;

  di->cachefd = __av_get_tmpfd(DUMMYV);
  if(di->cachefd == -1) 
    __av_log(AVLOG_WARNING, "Filtprog: Error opening cachefile [%i]", 
	     DUMMYV->errn);
  di->cacheptr = 0;

  return (void *) di;
}
                        
int __av_close_filtprog(ave *v, void *devinfo)
{
  int res = 0;
  filtdi *di = (filtdi *) devinfo;

  stop_filter(di);
  if(di->cachefd != -1) __av_localclose(DUMMYV, di->cachefd);
  __av_free(di);

  return res;
}
