/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_LINK module (vfs layer)
*/

#include "op_util.h"

#include <utime.h>


static LOCAL2(chmod, int, const char *, mode_t)
static LOCAL2(fchmod, int, int, mode_t)
static LOCAL3(chown, int, const char *, uid_t, gid_t)
static LOCAL3(fchown, int, int, uid_t, gid_t)

static int __av_localutime(ave *v, const char *path, struct utimbuf *buf)
{
  int res;
  
  errno = v->errn;
  res = __av_realutime(path, (void *) buf);
  v->errn = errno;
  return res;
}

#ifdef EMULATE_UTIMES
#include <sys/time.h>

static int __av_localutimes(ave *v, const char *path, struct timeval *tvp)
{
  int res;
  
  errno = v->errn;
  res = __av_realutime(path, (void *) tvp);
  v->errn = errno;
  return res;
}
#endif /* EMULATE_UTIMES */

#ifdef HAVE_LCHOWN
static LOCAL3(lchown, int, const char *, uid_t, gid_t)
#endif /* HAVE_LCHOWN */


int __av_chmod(ave *v, vpath *path, avmode_t mode)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localchmod(v, path->ipath, mode);
  else {
    if(vdev->chmod != NULL)
      res = (*vdev->chmod)(v, vdev->devdata, path, mode);
    else if(vdev->fchmod != NULL) {
      void *devinfo;
      
      devinfo = (*vdev->open)(v, vdev->devdata, path, 0, 0);
      if(devinfo == NULL) res = -1;
      else {
	res = (*vdev->fchmod)(v, devinfo, mode);
	(*vdev->close)(DUMMYV, devinfo);
      }
    } 
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtchmod(const char *path, mode_t mode)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("chmod", path, mode, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localchmod(&v, path, mode);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_chmod(&v, vp, mode);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (chmod)", path, res, v.errn);

  errno = v.errn;
  return res;
}


int __av_fchmod(ave *v, int fh, avmode_t mode)
{
  finfo *fi;
  int res;

  if(IS_LOCAL(fh)) {
    res = __av_localfchmod(v, fh, mode);
    return res;
  }

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;
  
  return (*fi->vdev->fchmod)(v, fi->devinfo, mode);
}


int __av_virtfchmod(int fh, mode_t mode)
{
  ave v;
  int res;

  v.errn = errno;
  if(IS_LOCAL(fh)) res = __av_localfchmod(&v, fh, mode);
  else res = __av_fchmod(&v, fh, mode);
  errno = v.errn;

  return res;
}


static int common_chown(ave *v, vpath *path, avuid_t uid, avgid_t gid,
			int deref)
{
  int res;
  struct vdev_info *vdev = path->vdev;

  if(vdev->chown != NULL)
    res = (*vdev->chown)(v, vdev->devdata, path, uid, gid, deref);
  else if(vdev->fchown != NULL) {
    void *devinfo;

    devinfo = (*vdev->open)(v, vdev->devdata, path, 0, 0);
    if(devinfo == NULL) res = -1;
    else {
      res = (*vdev->fchown)(v, devinfo, uid, gid);
      (*vdev->close)(DUMMYV, devinfo);
    }
  }
  else {
    v->errn = EPERM;
    res = -1;
  }

  return res;
}

int __av_chown(ave *v, vpath *path, avuid_t uid, avgid_t gid)
{
  int res;
  
  if(path->vdev == NULL) 
    res = __av_localchown(v, path->ipath, uid, gid);
  else 
    res = common_chown(v, path, uid, gid, 0);

  return res;
}


int __av_virtchown(const char *path, uid_t uid, gid_t gid)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("chown", path, uid, gid);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localchown(&v, path, uid, gid);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_chown(&v, vp, uid, gid);
      __av_free_vpath(vp);
    }
  }
  
  AVLOGSYS("  (chown)", path, res, v.errn);

  errno = v.errn;
  return res;
}


int __av_fchown(ave *v, int fh, avuid_t uid, avgid_t gid)
{
  finfo *fi;
  int res;

  if(IS_LOCAL(fh)) {
    res = __av_localfchown(v, fh, uid, gid);
    return res;
  }

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;
  
  return (*fi->vdev->fchown)(v, fi->devinfo, uid, gid);
}


int __av_virtfchown(int fh, uid_t uid, gid_t gid)
{
  ave v;
  int res;

  v.errn = errno;
  if(IS_LOCAL(fh)) res = __av_localfchown(&v, fh, uid, gid);
  else res = __av_fchown(&v, fh, uid, gid);
  errno = v.errn;

  return res;
}


#ifdef HAVE_LCHOWN

int __av_virtlchown(const char *path, uid_t uid, gid_t gid)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("lchown", path, uid, gid);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_locallchown(&v, path, uid, gid);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
  
      if(vp->vdev == NULL) 
	res = __av_locallchown(&v, vp->ipath, uid, gid);
      else 
	res = common_chown(&v, vp, uid, gid, 1);

      __av_free_vpath(vp);
    }
  }
  
  AVLOGSYS("  (lchown)", path, res, v.errn);

  errno = v.errn;
  return res;
}

#endif /* HAVE_LCHOWN */



int __av_utime(ave *v, vpath *path, struct avutimbuf *buf)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) {
    if(buf != NULL) {
      struct utimbuf lbuf;
      
      lbuf.actime = buf->actime;
      lbuf.modtime = buf->modtime;
      res = __av_localutime(v, path->ipath, &lbuf);
    }
    else res = __av_localutime(v, path->ipath, NULL);
  }
  else {
    if(vdev->utime != NULL)
      res = (*vdev->utime)(v, vdev->devdata, path, buf);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtutime(const char *path, void *lbuf)
{
  ave v;
  int res;
  int local;
  struct utimbuf *buf = (struct utimbuf *) lbuf;

  v.errn = errno;

  AVLOGSYS("utime", path, buf ? buf->actime : 0 , buf ? buf->modtime : 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localutime(&v, path, buf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_localutime(&v, vp->ipath, buf);
      else {
	if(buf != NULL) {
	  struct avutimbuf vbuf;
	  
	  vbuf.actime = buf->actime;
	  vbuf.modtime = buf->modtime;
	  res = __av_utime(&v, vp, &vbuf);
	}
	else res = __av_utime(&v, vp, NULL);
      }

      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (utime)", path, res, v.errn);

  errno = v.errn;
  return res;
}

#ifdef EMULATE_UTIMES

int __av_virtutimes(const char *path, void *lbuf)
{
  ave v;
  int res;
  int local;
  struct timeval *tvp = (struct timeval *) lbuf;

  v.errn = errno;

  AVLOGSYS("utimes", path, tvp ? tvp[0].tv_sec: 0, tvp ? tvp[1].tv_sec : 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localutimes(&v, path, tvp);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_localutimes(&v, vp->ipath, tvp);	
      else {
	if(tvp != NULL) {
	  struct avutimbuf vbuf;
	  
	  vbuf.actime  = tvp[0].tv_sec + (tvp[0].tv_usec / 1000000);
	  vbuf.modtime = tvp[1].tv_sec + (tvp[1].tv_usec / 1000000);
	  res = __av_utime(&v, vp, &vbuf);
	}
	else res = __av_utime(&v, vp, NULL);
      }

      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (utimes)", path, res, v.errn);

  errno = v.errn;
  return res;
}

#endif /* EMULATE_UTIMES */
