/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_LINK module (vfs layer)
*/

#include "op_util.h"

LOCAL1(unlink, int, const char *)
LOCAL2(mkdir, int, const char *, avmode_t)

static LOCAL1(rmdir, int, const char *)
static LOCAL2(symlink, int, const char *, const char *)
static LOCAL2(rename, int, const char *, const char *)
static LOCAL2(link, int, const char *, const char *)
static LOCAL3(mknod, int, const char *, mode_t, dev_t)

static int __av_unlink(ave *v, vpath *path)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localunlink(v, path->ipath);
  else {
    if(vdev->unlink != NULL)
      res = (*vdev->unlink)(v, vdev->devdata, path);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtunlink(const char *path)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("unlink", path, 0, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localunlink(&v, path);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      res = __av_unlink(&v, vp);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (unlink)", path, res, v.errn);

  errno = v.errn;
  return res;
}


static int __av_rmdir(ave *v, vpath *path)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localrmdir(v, path->ipath);
  else {
    if(vdev->rmdir != NULL)
      res = (*vdev->rmdir)(v, vdev->devdata, path);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtrmdir(const char *path)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("rmdir", path, 0, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localrmdir(&v, path);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      res = __av_rmdir(&v, vp);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (rmdir)", path, res, v.errn);

  errno = v.errn;
  return res;
}


static int __av_mkdir(ave *v, vpath *path, mode_t mode)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localmkdir(v, path->ipath, mode);
  else {
    if(vdev->mkdir != NULL)
      res = (*vdev->mkdir)(v, vdev->devdata, path, mode);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtmkdir(const char *path, mode_t mode)
{
  ave v;
  int res;
  int local;
  struct stat stbuf;

  v.errn = errno;

  AVLOGSYS("mkdir", path, mode, 0);

  local = 0; res = -1;
  if(maybe_local(path) && 
     (path_local(path) || __av_realstat(path, &stbuf) != -1)) {
    res = __av_localmkdir(&v, path, mode);
    local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      res = __av_mkdir(&v, vp, mode);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (mkdir)", path, res, v.errn);

  errno = v.errn;
  return res;
}


static int __av_symlink(ave *v, const char *path, vpath *newpath)
{
  int res;
  struct vdev_info *vdev = newpath->vdev;
  
  if(vdev == NULL) 
    res = __av_localsymlink(v, path, newpath->ipath);
  else {
    if(vdev->symlink != NULL)
      res = (*vdev->symlink)(v, vdev->devdata, path, newpath);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtsymlink(const char *path, const char *newpath)
{
  ave v;
  int res;
  int local;
  struct stat stbuf;

  v.errn = errno;

  AVLOGSYS("symlink", path, 0, 0);
  AVLOGSYS("symlink", newpath, 0, 0);

  local = 0; res = -1;
  if(maybe_local(newpath) && 
     (path_local(newpath) || __av_realstat(newpath, &stbuf) != -1)) {
    res = __av_localsymlink(&v, path, newpath);
    local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, newpath, 0);
    if(vp != NULL) {
      res = __av_symlink(&v, path, vp);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (symlink)", newpath, res, v.errn);

  errno = v.errn;
  return res;
}


static int same_file(vpath *vp1, vpath *vp2)
{
  struct stat st1;
  struct stat st2;
  
  if(vp1 == NULL && vp2 == NULL) return 1;
  if(vp1 == NULL || vp2 == NULL) return 0;

  if(vp1->vdev != vp2->vdev) return 0;
  if(vp1->vdev == NULL) {
    if(__av_realstat(vp1->ipath, &st1) == -1) return 0;
    if(__av_realstat(vp2->ipath, &st2) == -1) return 0;
    
    if(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino) return 1;
    return 0;
  }
  
  if(strcmp(vp1->ipath, vp2->ipath) != 0) return 0;
  
  return same_file(vp1->base, vp2->base);
}

static int samevdev(ave *v, vpath *vp1, vpath *vp2)
{
  if(vp1->vdev != vp2->vdev || !same_file(vp1->base, vp2->base)) {
    v->errn = EXDEV;
    return -1;
  }
  return 0;
}

static int __av_rename(ave *v, const char *path, vpath *newpath)
{
  int res;
  struct vdev_info *vdev = newpath->vdev;
  
  if(vdev == NULL) 
    res = __av_localrename(v, path, newpath->ipath);
  else {
    if(vdev->rename != NULL)
      res = (*vdev->rename)(v, vdev->devdata, path, newpath);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtrename(const char *path, const char *newpath)
{
  ave v;
  int res;
  int local;
  struct stat stbuf;

  v.errn = errno;

  AVLOGSYS("rename", path, 0, 0);
  AVLOGSYS("rename", newpath, 0, 0);

  local = 0; res = -1;
  if(maybe_local(path) && maybe_local(newpath) && 
     (path_local(newpath) || __av_realstat(newpath, &stbuf) != -1)) {
    res = __av_localrename(&v, path, newpath);
    local = 1;
  }
  
  if(!local) {
    vpath *vp;
    vpath *newvp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      newvp = __av_get_vpath(&v, newpath, 0);
      if(newvp != NULL) {
	res = samevdev(&v, vp, newvp);
	if(res != -1) res = __av_rename(&v, vp->ipath, newvp);
      }
      __av_free_vpath(newvp);
    }
    __av_free_vpath(vp);
  }

  AVLOGSYS("  (rename)", newpath, res, v.errn);

  errno = v.errn;
  return res;
}

static int __av_link(ave *v, const char *path, vpath *newpath)
{
  int res;
  struct vdev_info *vdev = newpath->vdev;
  
  if(vdev == NULL) 
    res = __av_locallink(v, path, newpath->ipath);
  else {
    if(vdev->link != NULL)
      res = (*vdev->link)(v, vdev->devdata, path, newpath);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}



int __av_virtlink(const char *path, const char *newpath)
{
  ave v;
  int res;
  int local;
  struct stat stbuf;

  v.errn = errno;

  AVLOGSYS("link", path, 0, 0);
  AVLOGSYS("link", newpath, 0, 0);

  local = 0; res = -1;
  if(maybe_local(path) && maybe_local(newpath) && 
     (path_local(newpath) || __av_realstat(newpath, &stbuf) != -1)) {
    res = __av_locallink(&v, path, newpath);
    local = 1;
  }
  
  if(!local) {
    vpath *vp;
    vpath *newvp;
    
    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      newvp = __av_get_vpath(&v, newpath, 0);
      if(newvp != NULL) {
	res = samevdev(&v, vp, newvp);
	if(res != -1) res = __av_link(&v, vp->ipath, newvp);
      }
      __av_free_vpath(newvp);
    }
    __av_free_vpath(vp);
  }

  AVLOGSYS("  (link)", newpath, res, v.errn);

  errno = v.errn;
  return res;
}


static int __av_mknod(ave *v, vpath *path, mode_t mode, dev_t dev)
{
  int res;
  struct vdev_info *vdev = path->vdev;
  
  if(vdev == NULL) 
    res = __av_localmknod(v, path->ipath, mode, dev);
  else {
    if(vdev->mknod != NULL)
      res = (*vdev->mknod)(v, vdev->devdata, path, mode, dev);
    else {
      v->errn = EPERM;
      res = -1;
    }
  }
  return res;
}


int __av_virtmknod(const char *path, mode_t mode, dev_t dev)
{
  ave v;
  int res;
  int local;
  struct stat stbuf;

  v.errn = errno;

  AVLOGSYS("mknod", path, mode, 0);

  local = 0; res = -1;
  if(maybe_local(path) && 
     (path_local(path) || __av_realstat(path, &stbuf) != -1)) {
    res = __av_localmknod(&v, path, mode, dev);
    local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      res = __av_mknod(&v, vp, mode, dev);
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (mknod)", path, res, v.errn);

  errno = v.errn;
  return res;
}
