/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    OP_STAT module (vfs layer)
*/

#include "op_util.h"

static void convert_stat(struct avstat *vbuf, struct stat *lbuf)
{
  lbuf->st_dev      = vbuf->dev;
  lbuf->st_ino      = vbuf->ino;
  lbuf->st_mode     = vbuf->mode;
  lbuf->st_nlink    = vbuf->nlink;
  lbuf->st_uid      = vbuf->uid;
  lbuf->st_gid      = vbuf->gid;
  lbuf->st_rdev     = vbuf->rdev;
  lbuf->st_size     = vbuf->size;
  lbuf->st_blksize  = vbuf->blksize;
  lbuf->st_blocks   = vbuf->blocks;
  lbuf->st_atime    = vbuf->atime;
  lbuf->st_mtime    = vbuf->mtime;
  lbuf->st_ctime    = vbuf->ctime;
}

static void convert_stat_back(struct stat *lbuf, struct avstat *vbuf)
{
  vbuf->dev      = lbuf->st_dev;
  vbuf->ino      = lbuf->st_ino;
  vbuf->mode     = lbuf->st_mode;
  vbuf->nlink    = lbuf->st_nlink;
  vbuf->uid      = lbuf->st_uid;
  vbuf->gid      = lbuf->st_gid;
  vbuf->rdev     = lbuf->st_rdev;
  vbuf->size     = lbuf->st_size;
  vbuf->blksize  = lbuf->st_blksize;
  vbuf->blocks   = lbuf->st_blocks;
  vbuf->atime    = lbuf->st_atime;
  vbuf->mtime    = lbuf->st_mtime;
  vbuf->ctime    = lbuf->st_ctime;
}

static LOCAL2(fstat, int, int, struct stat *)
static LOCAL2(stat, int, const char *, struct stat *)
static LOCAL2(lstat, int, const char *, struct stat *)

int __av_fstat(ave *v, int fh, struct avstat *buf, int flags)
{
  finfo *fi;
  int res;

  if(IS_LOCAL(fh)) {
    struct stat lbuf;
    res = __av_localfstat(v, fh, &lbuf);
    convert_stat_back(&lbuf, buf);
    return res;
  }

  fi = check_fh(v, fh);
  if(fi == NULL) return -1;

  return (*fi->vdev->fstat)(v, fi->devinfo, buf, flags);
}


int __av_virtfstat(int fh, void *lbuf)
{
  struct stat *buf = (struct stat *) lbuf;
  ave v;
  int res;

  v.errn = errno;
  if(IS_LOCAL(fh)) res = __av_localfstat(&v, fh, buf);
  else {
    struct avstat vbuf;
    res = __av_fstat(&v, fh, &vbuf, __av_priv.avfs_flags);
    convert_stat(&vbuf, buf);
  }
  errno = v.errn;
  return res;
}

static int common_stat(ave *v, vpath *path, struct avstat *buf, int flags, 
		       int deref)
{
  int res;
  struct vdev_info *vdev = path->vdev;

  if(vdev->stat != NULL)
    res =  (*vdev->stat)(v, vdev->devdata, path, buf, flags, deref);
  else if(vdev->fstat != NULL) {
    void *devinfo;

    devinfo = (*vdev->open)(v, vdev->devdata, path, 0, 0);
    if(devinfo == NULL) res = -1;
    else {
      res = (*vdev->fstat)(v, devinfo, buf, flags);
      (*vdev->close)(DUMMYV, devinfo);
    }
  }
  else {
    v->errn = ENODEV;
    res = -1;
  }

  return res;
}

int __av_stat(ave *v, vpath *path, struct avstat *buf, int flags)
{
  int res;
  
  if(path->vdev == NULL) {
    struct stat lbuf;

    res = __av_localstat(v, path->ipath, &lbuf);
    if(res != -1) convert_stat_back(&lbuf, buf);
  }
  else 
    res = common_stat(v, path, buf, flags, 1);

  return res;
}

int __av_virtstat(const char *path, void *lbuf)
{
  struct stat *buf = (struct stat *) lbuf;
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("stat", path, (int) buf, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localstat(&v, path, buf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_localstat(&v, vp->ipath, buf);
      else {
	struct avstat vbuf;
	
	res = common_stat(&v, vp, &vbuf, __av_priv.avfs_flags, 1);
	if(res != -1) convert_stat(&vbuf, buf);
      }
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (stat)", path, res, v.errn);

  errno = v.errn;
  return res;
}

int __av_virtlstat(const char *path, void *lbuf)
{
  struct stat *buf = (struct stat *) lbuf;
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("lstat", path, (int) buf, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_locallstat(&v, path, buf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_locallstat(&v, vp->ipath, buf);
      else {
	struct avstat vbuf;

	res = common_stat(&v, vp, &vbuf, __av_priv.avfs_flags, 0);
	if(res != -1) convert_stat(&vbuf, buf);
      }
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (lstat)", path, res, v.errn);

  errno = v.errn;
  return res;
}


#ifdef HAVE_OPEN64


struct  sun_stat64 {
  dev_t        st_dev;
  long         st_pad1[3];
  p_ino64_t    st_ino;
  mode_t       st_mode;
  nlink_t      st_nlink;
  uid_t        st_uid;
  gid_t        st_gid;
  dev_t        st_rdev;
  long         st_pad2[2];
  p_off64_t    st_size;
  timestruc_t  st_atim;
  timestruc_t  st_mtim;
  timestruc_t  st_ctim;
  long         st_blksize;
  p_blkcnt64_t st_blocks;
}; 

static void convert_stat64(struct avstat *vbuf, void *buf)
{
  struct sun_stat64 *lbuf = (struct sun_stat64 *) buf;
  
  lbuf->st_dev      = vbuf->dev;
  lbuf->st_ino      = vbuf->ino;
  lbuf->st_mode     = vbuf->mode;
  lbuf->st_nlink    = vbuf->nlink;
  lbuf->st_uid      = vbuf->uid;
  lbuf->st_gid      = vbuf->gid;
  lbuf->st_rdev     = vbuf->rdev;
  lbuf->st_size     = vbuf->size;
  lbuf->st_blksize  = vbuf->blksize;
  lbuf->st_blocks   = vbuf->blocks;
  lbuf->st_atime    = vbuf->atime;
  lbuf->st_mtime    = vbuf->mtime;
  lbuf->st_ctime    = vbuf->ctime;
}


static LOCAL2(stat64, int, const char *, void *)
static LOCAL2(lstat64, int, const char *, void *)
static LOCAL2(fstat64, int, int, void *)

int __av_virtstat64(const char *path, void *buf)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("stat64", path, (int) buf, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_localstat64(&v, path, buf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 1);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_localstat64(&v, vp->ipath, buf);
      else {
	struct avstat vbuf;
	
	res = common_stat(&v, vp, &vbuf, __av_priv.avfs_flags, 1);
	if(res != -1) convert_stat64(&vbuf, buf);
      }
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (stat64)", path, res, v.errn);

  errno = v.errn;
  return res;
}


int __av_virtlstat64(const char *path, void *buf)
{
  ave v;
  int res;
  int local;

  v.errn = errno;

  AVLOGSYS("lstat64", path, (int) buf, 0);

  local = 0; res = -1;
  if(maybe_local(path)) {
    res = __av_locallstat64(&v, path, buf);
    if(sure_local(&v, res, path)) local = 1;
  }
  
  if(!local) {
    vpath *vp;

    vp = __av_get_vpath(&v, path, 0);
    if(vp != NULL) {
      if(vp->vdev == NULL) 
	res = __av_locallstat64(&v, vp->ipath, buf);
      else {
	struct avstat vbuf;

	res = common_stat(&v, vp, &vbuf, __av_priv.avfs_flags, 0);
	if(res != -1) convert_stat64(&vbuf, buf);
      }
      __av_free_vpath(vp);
    }
  }

  AVLOGSYS("  (lstat64)", path, res, v.errn);

  errno = v.errn;
  return res;
}

int __av_virtfstat64(int fh, void *buf)
{
  ave v;
  int res;

  v.errn = errno;
  if(IS_LOCAL(fh)) res = __av_localfstat64(&v, fh, buf);
  else {
    struct avstat vbuf;
    res = __av_fstat(&v, fh, &vbuf, __av_priv.avfs_flags);
    convert_stat64(&vbuf, buf);
  }
  errno = v.errn;
  return res;
}

#endif /* HAVE_OPEN64 */


int __av_is_virtual(const char *path)
{
  char *pathv;
  ave v;
  int len;
  int res;
  int local;
  struct stat stbuf;
  
  v.errn = errno;
  local = 0; res = -1;
  
  len = strlen(path);

  pathv = __av_malloc(&v, len + 2);

  if(pathv == AVNULL) res = -1;
  else {
    strcpy(pathv, path);
    pathv[len] = VDEV_SEP_CHAR;
    pathv[len+1] = 0;
    
    if(maybe_local(pathv)) {
      res = __av_localstat(&v, pathv, &stbuf);
      if(sure_local(&v, res, pathv)) local = 1;
    }
    if(local) {
      v.errn = ENOENT;
      res = -1;
    }
    else {
      vpath *vp;
    
      vp = __av_get_vpath(&v, pathv, 0);
      if(vp != NULL) {
	if(vp->vdev == NULL) {
	  v.errn = ENOENT;
	  res = -1;
	}
	else res = (vp->vdev->flags & VDEV_FLAG_ISDIR);
	__av_free_vpath(vp);
      }
    }
    __av_free(pathv);
  }

  errno = v.errn;
  return res;
}
