/*
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    Path parser module
*/

#include "avfsint.h"
#include "resolve.h"

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

static int create_vpath(ave *v, workpath *path, vpath *base, vpath **vpp);

vpath *__av_get_base(vpath *vp)
{
  return vp->base;
}

const char *__av_get_extra(vpath *vp)
{
  return vp->ipath;
}

struct vdev_info *__av_get_vdev(vpath *vp)
{
  return vp->vdev;
}


static void destroy(void)
{
  int i;
  finfo *fi;

  if(!__av_priv.inited) return;

  fi = __av_priv.file_table;
  for(i = 0; i < AVFS_MAX_FILES; fi++, i++) {
    if(fi->type & FVIRTUAL) __av_close(DUMMYV, i);
  }

  /* Delete tmp-directory */

  if(__av_priv.tmpdir != AVNULL) 
    __av_unlink_recursive(__av_priv.tmpdir);

#if 0
  for(i = 0; i < __av_priv.vdev_num; i++) {
    __av_free(__av_priv.vdevs[i]->minors);
    if(__av_priv.vdevs[i]->destroy != NULL) 
      (*__av_priv.vdevs[i]->destroy)(__av_priv.vdevs[i]);
  }
  __av_free(__av_priv.vdevs);

  __av_priv.inited = 0;
#endif
}

extern int __av_init_module_avfsstat(ave *v);

static int init(ave *v)
{
  int i;
  static int started = 0;

  /* FIXME: Not nice sanity check */
  if(started) return 0;
  started = 1;

  AVLOGSYS("INIT", "", __av_priv.inited, 0);

  if(!__av_priv.vcwd_inited) __av_init_vcwd();

  __av_init_misc();

  for(i = 0; i < AVFS_MAX_FILES; i++) {
    __av_priv.file_table[i].type = 0;
    __av_priv.file_table[i].virtdir = 0;
    __av_priv.file_table[i].devinfo = NULL;
    __av_priv.file_table[i].path = NULL;
  }

  __av_priv.vdevs = NULL;
  __av_priv.vdevs_size = 0;
  __av_priv.vdev_num = 0;

  __av_init_module_avfsstat(DUMMYV);
  __av_init_modules();

  __av_priv.avfs_flags = 0;
  __av_priv.tmpdir = NULL;

  __av_priv.inited = 1;

  AVLOGSYS("  INIT", "", __av_priv.inited, 0);
  if(atexit(destroy) != 0) 
    __av_log(AVLOG_WARNING, "Could not register 'destroy' function"); 

  return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * A handler name can consist of alphanumeric characters and '_' This
 * function returns the handler pointer or NULL, if no handler could be
 * found
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static struct vdev_info *find_vdev_name(workpath *path)
{
  int beg;
  unsigned len;
  int vdevi, vi;

  beg = path->i;
  
  for(; !AR_END(path); AR_STEP(path)) {
    int ch = (unsigned char) AR_CURR(path);
    if(!isalpha(ch) && !isdigit(ch) && ch != '_') break;
  }
    
  len = path->i - beg;

  vdevi = -1;
  if(len != 0) 
    for(vi = 0; vi < __av_priv.vdev_num; vi++)
      if(__av_priv.vdevs[vi]->name != AVNULL && 
	 strlen(__av_priv.vdevs[vi]->name) == len &&
	 strncmp(path->s+beg, __av_priv.vdevs[vi]->name, len) == 0) {
	vdevi = vi;
	break;
      }
  
  return (vdevi != -1) ? __av_priv.vdevs[vdevi] : AVNULL;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Resolves '.'s '..'s and multiple '/'s in the path. It returns 1 on
 * successful resolvation, or 0 if the '..'s lead out of this path segment.
 * Returns -1 on error (pathname too long)
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int resolve_path(ave *v, workpath *path, char *ipath, int *top)
{
  int to = *top;
  int segbeg;
  int seglen;

  segbeg = to;
  while(1) {
    if(AR_END(path) || AR_CURR(path) == DIR_SEP_CHAR) {
      seglen = to - segbeg;

      /* The path always begins with a '/'. Everything before the 
	 first '/' is the handler's parameters */
      if(segbeg > 0) {
	  
	/* '.' and '..' are special */
	if(ipath[segbeg] == '.') {
	  if(seglen == 1) to = segbeg - 1;
	  if(seglen == 2 && ipath[segbeg+1] == '.') {
	    to = segbeg - 2;
	    for(; to >= 0 && ipath[to] != DIR_SEP_CHAR; to--);
	    if(to < 0) return 0;
	  }
	}
      }
      while(AR_CURR(path) == DIR_SEP_CHAR) AR_STEP(path);
      if(AR_END(path)) break;
      ipath[to++] = DIR_SEP_CHAR;
      segbeg = to;
    }
    else {
      if(!AR_CURR(path)) break;

      ipath[to++] = AR_CURR(path);
      AR_STEP(path);
    }
    if(checkl(v, to)) return -1;
  }

  *top = to;

  return 1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Analyzes the next path segment (after the handler name). Returns -1 on
 * error, otherwise returns 1. If the analysis of the path was complete
 * *vpp will contain the whole vpath, else *vpp will remain NULL.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int create_segment(ave *v, workpath *path, vpath *newvp, vpath **vpp)
{
  char ipath[AVFS_MAX_PATHLEN];
  int islocal, res;
  int to;
  
  islocal = (newvp->base == AVNULL);
  newvp->ipath = ipath;

  to = 0;
  do {
    if(islocal) 
      while(AR_CURR(path)) {
	ipath[to++] = AR_CURR(path);
	if(checkl(v, to)) return -1;
	AR_STEP(path);
      }
    else {
      if(newvp->vdev->resolve != AVNULL) {
	ipath[to] = '\0';
	res = (*newvp->vdev->resolve)(v, newvp->vdev->devdata, newvp, 
				      ipath, &to, path);
      }
      else 
	res = resolve_path(v, path, ipath, &to);

      if(res == -1) return -1;
      else if(!res) return 1;
    }
    
    if(!AR_END(path)) {
      ipath[to] = '\0';
      res = create_vpath(v, path, newvp, vpp);
      if(res == -1) return -1;

      /* zero return value means, the handler could not be found */
      else if(res == 0) {
	if(!islocal) {
	  /* We don't allow nonexisting handlers inside virtual paths */
	  v->errn = ENOENT;
	  return -1;
	}
	ipath[to++] = VDEV_SEP_CHAR;
	if(checkl(v, to)) return -1;
	AR_STEP(path);
      }
      else {
	/* If the resolved path points to the parent directory: */
	if(*vpp == AVNULL) {
	  to--;
	  while(to >= 0 && ipath[to] == DIR_SEP_CHAR) to--;
	  while(to >= 0 && ipath[to] != DIR_SEP_CHAR) to--;
	  
	  if(islocal) {
	    to++;
	    ipath[to++] = '.';
	    if(checkl(v, to)) return -1;
    	  }
	  /* We have stepped out from this handler */
	  else if(to < 0) return 1;
	}
      }
    }
  } while(!AR_END(path));

  ipath[to] = '\0';
  
  newvp->ipath = __av_strdup(v, ipath);
  if(newvp->ipath == AVNULL) return -1;
  
  if(*vpp == AVNULL) *vpp = newvp;
  return 1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Tries to find the handler based on the extension of the base-file.
 * Return values:
 *   -1:  Error,  
 *    0:  No handler could be found, for the file extension
 *    1:  Otherwise,  *vpp indicates, whether the analysis was complete 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int create_auto_vpath(ave *v, workpath *path, const char *file, 
			     int filelen, vpath *base, vpath **vpp)
{
  int res;
  int vi, ei;
  struct ext_info *exts;
  vpath *newvp;
  
  for(vi = 0; vi < __av_priv.vdev_num; vi++) {
    exts = __av_priv.vdevs[vi]->exts;
    if(exts != AVNULL) 
      for(ei = 0; exts[ei].from != AVNULL; ei++) {
	int extlen = strlen(exts[ei].from);
	
	if(filelen >= extlen && 
	   strncmp(file+filelen-extlen, exts[ei].from, extlen) == 0) {
	  char *ipath;

	  AV_NEW(v, newvp);
	  ipath = __av_malloc(v, 1);
	  if(newvp == AVNULL || ipath == AVNULL) {
	    __av_free(newvp);
	    __av_free(ipath);
	    return -1;
	  }
	  ipath[0] = '\0';
	  
	  newvp->vdev = __av_priv.vdevs[vi];
	  newvp->base = base;
	  newvp->av = 1;
	  newvp->ipath = ipath;

	  if(exts[ei].to == AVNULL) 
	    res = create_auto_vpath(v, path, file, filelen-extlen, newvp, vpp);
	  else {
	    int newextlen, newfilelen;
	    char *newfile;

	    newextlen = strlen(exts[ei].to);
	    newfilelen = filelen - extlen + newextlen;
	    
	    newfile = __av_malloc(v, newfilelen+1);
	    if(newfile == AVNULL) res = -1;
	    else {
	      strncpy(newfile, file, filelen - extlen);
	      strncpy(newfile + filelen - extlen, exts[ei].to, newextlen);
	      
	      res = create_auto_vpath(v, path, newfile, newfilelen, 
				      newvp, vpp);
	      __av_free(newfile);
	    }
	  }

	  if(res == 0) {
	    __av_free(ipath);
	    res = create_segment(v, path, newvp, vpp);
	    if(*vpp == AVNULL) __av_free(newvp);
	    return res;
	  }
	  
	  if(*vpp == AVNULL) {
	    __av_free(newvp);
	    __av_free(ipath);
	  }
	  return res;
	}
      }
  }
  return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Creates vpath from path. 
 *  Return values:
 *   -1:  Error,  
 *    0:  No handler could be found.
 *    1:  Otherwise,  *vpp indicates, whether the analysis was complete
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int create_vpath(ave *v, workpath *path, vpath *base, vpath **vpp)
{
  struct vdev_info *vdev;
  int res;
  vpath *newvp;

  /* If not local segment */
  if(base != AVNULL) {
    int atsave = path->i;

    AR_STEP(path);
    if(AR_END(path) || AR_CURR(path) == DIR_SEP_CHAR) {
      int ls;
      
      for(ls = path->i - 2; ls >= 0 && path->s[ls] != DIR_SEP_CHAR; ls--);
      ls++;
      return create_auto_vpath(v, path, path->s+ls, (path->i-1)-ls, base, vpp);
    }
      
    vdev = find_vdev_name(path);
    
    /* returning 0 means: could not resolve handler name */
    if(vdev == AVNULL) {
      path->i = atsave;
      return 0;
    }
  }
  else vdev = AVNULL;

  AV_NEW(v, newvp);
  if(newvp == AVNULL) return -1;
  newvp->vdev = vdev;
  newvp->base = base;
  newvp->av = 0;
  
  res = create_segment(v, path, newvp, vpp);
  if(*vpp == AVNULL) __av_free(newvp);
  return res;
}

static int preprocess_path(char *path)
{
  char *dst;
  int len;
  
  len = 0;
  for(dst = path; *path; dst++, len++, path++) {
    if(*path == VDEV_SEP_CHAR) {
      if(*(path+1) == VDEV_SEP_CHAR) *dst = *path++;
      else *dst = '\0';
    }
    else *dst = *path;
  }
  *dst = '\0';
  return len;
}

static char *full_path(ave *v, const char *origpath)
{
  char *path;
  int newpathlen;
  int addcwd;
  int at;

  if(origpath == AVNULL) return 0;

  newpathlen = strlen(origpath);
  if(origpath[0] != DIR_SEP_CHAR && __av_priv.vcwd[0]) {
    addcwd = 1;
    newpathlen += (1 + strlen(__av_priv.vcwd));
  }
  else addcwd = 0;

  path = __av_malloc(v, newpathlen + 1); 
  if(path == AVNULL) return AVNULL;

  at = 0;
  if(addcwd) {
    strcpy(path, __av_priv.vcwd);
    at = strlen(path);
    path[at++] = DIR_SEP_CHAR;
  }
  strcpy(path+at, origpath);

  return path;
}

vpath *__av_get_vpath(ave *v, const char *path, int resolvelast)
{
  char *newpath;
  vpath *vp = AVNULL;
  workpath wpath;
  int res;
  int len;

  if(!__av_priv.inited && init(v) == -1) return AVNULL;

  newpath = full_path(v, path);
  if(newpath == AVNULL) return AVNULL;

  len = preprocess_path(newpath);
  wpath.s = newpath;
  wpath.l = len;
  wpath.i = 0;
  wpath.resolvelast = resolvelast;
  res = create_vpath(v, &wpath, AVNULL, &vp);
  __av_free(newpath);
  
  if(res == -1) return AVNULL;
  
  return vp;
}

void __av_free_vpath(vpath *vp)
{
  if(vp != AVNULL) {
    __av_free_vpath(vp->base);
    __av_free(vp->ipath);
    
    __av_free(vp);
  }
}


static int ipath_len(const char *s)
{
  int cnt;

  for(cnt = 0; *s; s++, cnt++) 
    if(*s == VDEV_SEP_CHAR) cnt++;

  return cnt;
}

static void ipath_copy(char *dst, const char *src)
{
  for(; *src; dst++, src++) {
    *dst = *src;
    if(*src == VDEV_SEP_CHAR) *++dst = VDEV_SEP_CHAR;
  }
}

int __av_generate_path(ave *v, vpath *path, char **rpathp)
{
  char *rpath;
  int plen, at, lastat = 0, len;
  vpath *pp;

  plen = 0;
  for(pp = path; pp != AVNULL; pp = pp->base) {
    plen += ipath_len(pp->ipath);
    if(pp->vdev != AVNULL) plen += 1 + strlen(pp->vdev->name);
  }
  
  rpath = __av_malloc(v, plen + 1);
  if(rpath == AVNULL) return -1;

  at = plen;
  rpath[at] = '\0';
  for(pp = path; pp != AVNULL; pp = pp->base) {
    lastat = at;
    len = ipath_len(pp->ipath);
    at -= len;
    ipath_copy(rpath + at, pp->ipath);
    
    if(pp->vdev != AVNULL) {
      len = strlen(pp->vdev->name);
      at -= len;
      strncpy(rpath + at, pp->vdev->name, len);
      rpath[--at] = VDEV_SEP_CHAR;
    }
  }

  *rpathp  = rpath;

  return lastat;
}

#ifdef PARSE_DEBUG

#include <stdio.h>

int main(int argc, char *argv[])
{
  char *path;
  ave v;
  vpath *vp;
  vpath *vpx;
  char *newpath;
  int len;
  int i;
  int resolvelast;

  if(argc < 2) {
    fprintf(stderr, "usage: parse virtual_path\n");
    return 1;
  }
  path = argv[1];

  newpath = __av_strdup(DUMMYV, path);
  len = preprocess_path(newpath);
  
  printf("preprocessed path: <");
  for(i = 0; i < len; i++) {
    if(newpath[i]) putchar(newpath[i]);
    else putchar('|');
  }
  printf(">\n");

  for(resolvelast = 0; resolvelast <= 1; resolvelast++) {
    printf("------------ resolvelast = %i ------------\n", resolvelast);

    vp = __av_get_vpath(&v, path, resolvelast);
    
    if(vp == NULL) 
      printf("Error: %s\n", strerror(v.errn));
    
    
    for(vpx = vp; vpx != NULL; vpx = vpx->base) {
      printf("%s: <%s> (%c)\n",
	     vpx->vdev == NULL ? "local" : vpx->vdev->name,
	     vpx->ipath == NULL ? "(null)" : vpx->ipath,
	     vpx->av ? 'a' : 'n');
    }
    __av_free_vpath(vp);
  }
}

#endif
