/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)

    Based on the GNU tar sources (C) Free Software Foundation
    
    This file can be distributed under the GNU GPL. 
    See the file COPYING. 

    TAR module
*/

#include "gtar.h"
#include "archives.h"
#include "vdev.h"

#include <ctype.h>
#include <string.h>

/* Some constants from POSIX are given names.  */
#define NAME_FIELD_SIZE   100
#define PREFIX_FIELD_SIZE 155
#define UNAME_FIELD_SIZE   32
#define GNAME_FIELD_SIZE   32

struct tar_archdat {
  enum archive_format format;  
  int broken;
};

struct tar_header {
  char *name;
  char *linkname;
  avoff_t size;
};

struct sp_array
{
  avoff_t offset;
  int numbytes;
};

struct tar_fdidat {
  struct sp_array *sparsearray;
  int sp_array_len;
  avoff_t headeroff;
};

#define ISSPACE(x) isspace(x)
#define ISODIGIT(x) ((x) >= '0' && (x) < '8')

/*------------------------------------------------------------------------.
| Quick and dirty octal conversion.  Result is -1 if the field is invalid |
| (all blank, or nonoctal).						  |
`------------------------------------------------------------------------*/

static long from_oct(int digs, char *where)
{
  long value;

  while (ISSPACE ((int) *where))
    {				/* skip spaces */
      where++;
      if (--digs <= 0)
	return -1;		/* all blank field */
    }
  value = 0;
  while (digs > 0 && ISODIGIT (*where))
    {
      /* Scan til nonoctal.  */

      value = (value << 3) | (*where++ - '0');
      --digs;
    }

  if (digs > 0 && *where && !ISSPACE ((int) *where))
    return -1;			/* ended on non-space/nul */

  return value;
}

static int find_next_block(ave *v, struct arch_file *file, union block *blk)
{
  int res;

  res = __av_read(v, file->fh, blk->buffer, BLOCKSIZE);
  if(res == 0) return 0;
  if(res == -1) return -2;
  if(res < BLOCKSIZE) {
    v->errn = EIO;
    return -2;
  }
  
  file->ptr += BLOCKSIZE;
  return 1;
}


/* return values: -2 fatal, -1 bad header, 0 eof, 1 good */

static int read_header(ave *v, arch_file *file, union block *header, 
		       struct tar_header *hinf)
{
  int i;
  long unsigned_sum;		/* the POSIX one :-) */
  long signed_sum;		/* the Sun one :-( */
  long recorded_sum;
  char *p;
  char **longp;
  char *bp;
  union block data_block;
  int size, written;
  int res;
  char *next_long_name = AVNULL, *next_long_link = AVNULL;

  while (1)
    {
      res = find_next_block(v, file, header);
      if (res <= 0) break;  /* HEADER_END_OF_FILE */

      recorded_sum
	= from_oct (sizeof header->header.chksum, header->header.chksum);

      unsigned_sum = 0;
      signed_sum = 0;
      p = header->buffer;
      for (i = sizeof (*header); --i >= 0;)
	{
	  /* We can't use unsigned char here because of old compilers,
	     e.g. V7.  */

	  unsigned_sum += 0xFF & *p;
	  signed_sum += *p++;
	}

      /* Adjust checksum to count the "chksum" field as blanks.  */

      for (i = sizeof (header->header.chksum); --i >= 0;)
	{
	  unsigned_sum -= 0xFF & header->header.chksum[i];
	  signed_sum -= header->header.chksum[i];
	}
      unsigned_sum += ' ' * sizeof header->header.chksum;
      signed_sum += ' ' * sizeof header->header.chksum;

      if (unsigned_sum == sizeof header->header.chksum * ' ')
	{
	  /* This is a zeroed block...whole block is 0's except for the
	     blanks we faked for the checksum field.  */

	  res = 0;
	  break; /* HEADER_ZERO_BLOCK */
	}

      if (unsigned_sum != recorded_sum && signed_sum != recorded_sum) {
	res = -1;
	v->errn = EIO;
	break; /* HEADER_FAILURE */
      }

      /* Good block.  Decode file size and return.  */

      if (header->header.typeflag == LNKTYPE)
	hinf->size  = 0;	/* links 0 size on tape */
      else
	hinf->size = from_oct (1 + 12, header->header.size);

      header->header.name[NAME_FIELD_SIZE - 1] = '\0';
      if (header->header.typeflag == GNUTYPE_LONGNAME
	  || header->header.typeflag == GNUTYPE_LONGLINK)
	{
	  longp = ((header->header.typeflag == GNUTYPE_LONGNAME)
		   ? &next_long_name
		   : &next_long_link);

	  if (*longp) __av_free (*longp);
	  bp = *longp = (char *) __av_malloc (v, (avsize_t) hinf->size);
	  if (bp == AVNULL) {
	    res = -2;
	    break;
	  }

	  for (size = hinf->size; size > 0; size -= written)
	    {
	      res = find_next_block (v, file, &data_block);
	      if (res <= 0) break;
	      written = BLOCKSIZE;
	      if (written > size)
		written = size;

	      memcpy (bp, data_block.buffer, (avsize_t) written);
	      bp += written;
	    }
	  if(res <= 0) {
	    res = -2;
	    break;
	  }

	  /* Loop!  */

	}
      else
	{
	  hinf->name = __av_strdup (v, next_long_name ? next_long_name
				    : header->header.name);
	  hinf->linkname = __av_strdup (v, next_long_link ? next_long_link
					: header->header.linkname);

	  if(hinf->name == AVNULL || hinf->linkname == AVNULL) {
	    __av_free(hinf->name);
	    __av_free(hinf->linkname);
	    res = -2;
	  }
	  else res = 1;
	  break; /* HEADER_SUCCESS */
	}
    }

  __av_free(next_long_name);
  __av_free(next_long_link);
  return res;
}


static void decode_header (union block *header, struct avstat *stat_info,
			   enum archive_format *format_pointer)
{
  enum archive_format format;

  if (strcmp (header->header.magic, TMAGIC) == 0)
    format = POSIX_FORMAT;
  else if (strcmp (header->header.magic, OLDGNU_MAGIC) == 0)
    format = OLDGNU_FORMAT;
  else
    format = V7_FORMAT;
  *format_pointer = format;

  stat_info->mode = from_oct (8, header->header.mode);
  stat_info->mode &= 07777;
  stat_info->mtime = from_oct (1 + 12, header->header.mtime);

  if(header->header.typeflag == GNUTYPE_SPARSE)
    stat_info->size = 
      from_oct (1 + 12, header->oldgnu_header.realsize);
  else stat_info->size = from_oct (1 + 12, header->header.size);
  
  switch(header->header.typeflag) {
  case GNUTYPE_SPARSE:
  case REGTYPE:
  case AREGTYPE:
  case LNKTYPE:
  case CONTTYPE:
    stat_info->mode |= AV_IFREG;
    break;

  case GNUTYPE_DUMPDIR:
  case DIRTYPE:
    stat_info->mode |= AV_IFDIR;
    break;

  case SYMTYPE:
    stat_info->mode |= AV_IFLNK;
    break;
    
  case BLKTYPE:
    stat_info->mode |= AV_IFBLK;
    break;

  case CHRTYPE:
    stat_info->mode |= AV_IFCHR;
    break;

  case FIFOTYPE:
    stat_info->mode |= AV_IFIFO;
    break;
  }

  if (format == V7_FORMAT)
    {
      stat_info->uid = from_oct (8, header->header.uid);
      stat_info->gid = from_oct (8, header->header.gid);
      stat_info->rdev = 0;
    }
  else
    {
/* FIXME:
       if (numeric_owner_option
	  || !*header->header.uname
	  || !uname_to_uid (header->header.uname, &stat_info->uid))
*/
	stat_info->uid = from_oct (8, header->header.uid);
      
/* FIXME:
       if (numeric_owner_option
	  || !*header->header.gname
	  || !gname_to_gid (header->header.gname, &stat_info->gid))
*/
	stat_info->gid = from_oct (8, header->header.gid);
    
      switch (header->header.typeflag)
	{
	case BLKTYPE:
	case CHRTYPE:
	  stat_info->rdev = 
	    __av_mkdev (from_oct (8, header->header.devmajor),
			from_oct (8, header->header.devminor));
	  break;

	default:
	  stat_info->rdev = 0;
	}
    }
}


/* insert_tarentry() returns -2 on fatal, -1 on minor, 0 on OK */

static int insert_tarentry(ave *v, archive *arch, struct arch_file *file, 
			   union block *header, struct tar_header *hinf)
{
  struct avstat tarstat;
  struct tar_archdat *tad = (struct tar_archdat *) arch->udata;
  arch_entry *ent = AVNULL, *entlink = AVNULL;
  arch_inode *parentdir, *ino;
  char *name = AVNULL;
  
  __av_default_stat(&tarstat);

  decode_header(header, &tarstat, &tad->format);
  
  if(!hinf->name[0]) {
    __av_log(AVLOG_WARNING, "utar: Error - empty pathname");
    return -1;
  }

  /* Appears to be a file.  But BSD tar uses the convention that a slash
     suffix means a directory.  */
    
  if(AV_ISREG(tarstat.mode) && 
     hinf->name[strlen(hinf->name)-1] == DIR_SEP_CHAR) 
    tarstat.mode = (tarstat.mode & 07777) | AV_IFDIR;

  /* Check if already exists */
    
  ent = __av_find_entry(DUMMYV, arch->root, hinf->name);
  if(ent != AVNULL) {
    if(AV_ISDIR(ent->ino->st.mode)) {
      if(AV_ISDIR(tarstat.mode)) {
	ent->ino->st.mode = tarstat.mode;
	ent->ino->st.uid = tarstat.uid;
	ent->ino->st.gid = tarstat.gid;
	ent->ino->st.mtime = tarstat.mtime;
	ent->ino->origst = ent->ino->st;
	return 0;
      }
      else {
	__av_log(AVLOG_WARNING, 
		 "utar: Error - overwriting directory with file");
	return -1;
      }
    }
    else __av_unlink_entry(ent);
  }
  
  if(header->header.typeflag == LNKTYPE) {
    entlink = __av_find_entry(DUMMYV, arch->root, hinf->linkname);
    if(entlink == AVNULL) {
      __av_log(AVLOG_WARNING, "utar: Error - illegal hard link");
      return -1; 
    }
  }

  name = __av_split_path(v, hinf->name);
  if(name == AVNULL) goto error;
  if(!name[0]) {
    __av_log(AVLOG_WARNING, "utar: Error - empty filename");
    return -1;
  }
  parentdir = __av_make_dir_rec(v, arch, arch->root, hinf->name, 0);
  if(parentdir == AVNULL) goto error;

  ent = __av_new_entry(v);
  if(ent == AVNULL) goto error;

  if(header->header.typeflag == LNKTYPE) ino = entlink->ino;
  else {
    tarstat.blocks = AV_DIV(hinf->size, BLOCKSIZE);
    tarstat.blksize = BLOCKSIZE;
    tarstat.dev = arch->dev;
    tarstat.ino = arch->inoctr++;
    tarstat.atime = tarstat.mtime;  /* FIXME */
    tarstat.ctime = tarstat.mtime;

    if(AV_ISDIR(tarstat.mode)) ino = __av_make_dir(v, parentdir, &tarstat);
    else ino = __av_make_node(v, &tarstat);
    
    if(ino == AVNULL) goto error;
    
    ino->offset = file->ptr;
    ino->realsize = hinf->size;
    ino->typeflag = header->header.typeflag;

    if(header->header.typeflag == SYMTYPE) {
      ino->syml = hinf->linkname;
      hinf->linkname = AVNULL;
    }
  }

  ent->name = name;
  ent->ino = ino;
  __av_insert_entry(parentdir, ent);
  return 0;

 error:
  __av_free(name);
  __av_free(ent);
  return -2;
}

static int read_tarfile(ave *v, arch_file *file, archive *arch)
{
  union block header;
  struct tar_archdat *tad;
  static union block blck;
  struct tar_header hinf;
  int res;
  long size;
  avoff_t sres;

  AV_NEW(v, tad);
  if(tad == AVNULL) return -1;
  arch->udata = (void *) tad;

  tad->broken = 0;
  while(1) {
    res = read_header(v, file, &header, &hinf);
    if(res == -2) return -1;
    else if(res == -1) {
      tad->broken = 1;
      continue;
    }
    else if(res == 0) return 0;

    res = insert_tarentry(v, arch, file, &header, &hinf);
    __av_free(hinf.name);
    __av_free(hinf.linkname);

    if(res == -2) return -1;

    if (header.oldgnu_header.isextended) {
      do {
	res = find_next_block (v, file, &blck);
	if(res <= 0) break;
      }
      while(blck.sparse_header.isextended);
    }
    size = hinf.size;
    
    sres = __av_lseek(v, file->fh, AV_DIV(size, BLOCKSIZE) * BLOCKSIZE, 
		      AVSEEK_CUR);
    if(sres == -1) return -1;
    file->ptr = sres;
  }
}

static int parse_tarfile(ave *v, vpath *path, archive *arch)
{
  int res;
  arch_file file;
  const char *params;

  /* No parameters are allowed */
  params = EXTRA(path);
  if(params[0] && params[0] != DIR_SEP_CHAR) {
    v->errn = ENOENT;
    return -1;
  }
  
  file.fh = __av_open(v, BASE(path), AVO_RDONLY, 0);
  if(file.fh == -1) return -1;
  file.ptr = 0;

  res = read_tarfile(v, &file, arch);
  
  __av_close(DUMMYV, file.fh);
    
  return res;  
}


static int tar_close(ave *v, void *devinfo)
{
  arch_fdi *di = (arch_fdi *) devinfo;
  struct tar_fdidat *tfd = (struct tar_fdidat *) di->udata;
  
  __av_free(tfd->sparsearray);
  
  return (*di->vdev->close)(v, devinfo);
}


static void *tar_open(ave *v, void *devdata, vpath *path, int flags, int mode)
{
  arch_fdi *di;
  arch_devd *dd = (arch_devd *) devdata;
  struct tar_fdidat *tfd;

  AV_NEW(v, tfd);
  if(tfd == AVNULL) return AVNULL;

  di = (arch_fdi *) (*dd->vdev->open)(v, devdata, path, flags, mode);
  if(di == AVNULL) {
    __av_free(tfd);
    return AVNULL;
  }
  
  tfd->sparsearray = AVNULL;
  tfd->headeroff = di->ino->offset - BLOCKSIZE;

  di->udata = (void *) tfd;

  return (void *) di;
}

static int read_sparsearray(ave *v, arch_fdi *di)
{
  int res;
  union block header;
  int counter;
  struct sp_array *newarray, *sparses;
  struct tar_fdidat *tfd = (struct tar_fdidat *) di->udata;
  int size, len;
  
  di->file.ptr = __av_lseek(v, di->file.fh, tfd->headeroff, AVSEEK_SET);
  if(di->file.ptr == -1) return -1;
  
  res = find_next_block(v, &di->file, &header);
  if(res <= 0) return -1;

  size = 10;
  len = 0;
  sparses = (struct sp_array *) __av_malloc(v, size * sizeof(struct sp_array));
  if(sparses == AVNULL) return -1;

  for (counter = 0; counter < SPARSES_IN_OLDGNU_HEADER; counter++) {
    sparses[len].offset = 
      from_oct (1 + 12, header.oldgnu_header.sp[counter].offset);
    sparses[len].numbytes =
      from_oct (1 + 12, header.oldgnu_header.sp[counter].numbytes);

    if (!sparses[counter].numbytes) break;

    len++;
  }

  if (header.oldgnu_header.isextended)	{
    /* Read in the list of extended headers and translate them into
       the sparsearray as before.  */

    while (1) {
      res = find_next_block(v, &di->file, &header);
      if(res <= 0) goto error;
      
      for (counter = 0; counter < SPARSES_IN_SPARSE_HEADER; counter++) {
	if (counter + len > size - 1) {
	  /* Realloc the scratch area since we've run out of
	     room.  */

	  size *= 2;
	  newarray = (struct sp_array *)
	    __av_realloc (v, sparses, size * sizeof (struct sp_array));
	  if(newarray == AVNULL) goto error;
	  sparses = newarray;
	}

	if (header.sparse_header.sp[counter].numbytes[0] == 0) break;
	
	sparses[len].offset =
	  from_oct (1 + 12, header.sparse_header.sp[counter].offset);
	sparses[len].numbytes =
	  from_oct (1 + 12, header.sparse_header.sp[counter].numbytes);

	len++;
      }
      if (!header.sparse_header.isextended) break;
    }
  }
  
  tfd->sparsearray = sparses;
  tfd->sp_array_len = len;
  di->ino->offset = di->file.ptr; /* the correct offset */
  return 0;

 error:
  __av_free(sparses);
  return -1;
}

static avssize_t read_sparse(ave *v, arch_fdi *di, char *buf, avsize_t nbyte)
{
  struct tar_fdidat *tfd = (struct tar_fdidat *) di->udata;
  avoff_t offset   = di->ino->offset;
  avoff_t size     = di->ino->st.size;
  avoff_t realsize = di->ino->realsize;
  struct sp_array *sparses;
  avoff_t realoff;
  int ctr;
  avsize_t nact;
  avoff_t start, end;
  avoff_t spstart, spend;
  avoff_t cmstart, cmend;
  int res;

  if(AV_ISDIR(di->ino->st.mode)) {
    v->errn = EISDIR;
    return -1;
  }
  
  if(tfd->sparsearray == AVNULL) {
    res = read_sparsearray(v, di);
    if(res == -1) return -1;
  }
  sparses = tfd->sparsearray;

  if(di->ptr >= size) return 0;

  nact = AV_MIN(nbyte, (avsize_t) (size - di->ptr));
  start = di->ptr;
  end = start + nact;

  memset(buf, 0, nact);
  
  realoff = 0;
  ctr = 0; 
  while(ctr < tfd->sp_array_len && realoff < realsize) {
    spstart = sparses[ctr].offset;
    spend = spstart + sparses[ctr].numbytes;

    if(spstart < end && spend > start) {
      cmstart = AV_MAX(spstart, start);
      cmend   = AV_MIN(spend,   end);

      di->file.ptr = 
	__av_lseek(v, di->file.fh, realoff + offset + (cmstart - spstart), 
		  AVSEEK_SET);
      if(di->file.ptr == -1) return -1;
      res = __av_read(v, di->file.fh, buf + (cmstart - start), 
		      cmend - cmstart);
      if(res == -1) {
	di->file.ptr = -1;
	return -1;
      }
      di->file.ptr += res;
      if(res != (cmend - cmstart)) {
	v->errn = EIO;
	return -1;
      }
    }
    
    realoff += ((spend - spstart - 1) / BLOCKSIZE + 1) * BLOCKSIZE;
    ctr++;
  }
  
  di->ptr += nact;
  return nact;
}


static avssize_t tar_read(ave *v, void *devinfo, char *buf, avsize_t nbyte)
{
  arch_fdi *di = (arch_fdi *) devinfo;

  if(di->ino->typeflag == GNUTYPE_SPARSE) 
    return read_sparse(v, di, buf, nbyte);

  return (*di->vdev->read)(v, devinfo, buf, nbyte);
}

static int tar_fgetfile(ave *v, void *devinfo, avoff_t offset, avsize_t *lenp,
			avoff_t *retoffp)
{
  arch_fdi *di = (arch_fdi *) devinfo;
  
  if(di->ino->typeflag == GNUTYPE_SPARSE) {
    v->errn = ENODEV;
    return -1;
  }

  return (*di->vdev->fgetfile) (v, devinfo, offset, lenp, retoffp);
}


static struct ext_info tarexts[] = 
{
  { ".tar", AVNULL },
  { AVNULL,   AVNULL }
};

extern int __av_init_module_utar(ave *v);

int __av_init_module_utar(ave *v)
{
  struct vdev_info *tar_vdev;
  struct arch_devd *dd;

  tar_vdev = __av_init_arch(v, "utar");
  if(tar_vdev == AVNULL) return -1;

  dd = (arch_devd *) tar_vdev->devdata;

  dd->parsefunc       = parse_tarfile;

  tar_vdev->open      = tar_open;
  tar_vdev->close     = tar_close;
  tar_vdev->read      = tar_read;
  tar_vdev->fgetfile  = tar_fgetfile;

  tar_vdev->exts = tarexts;
  
  dd->major = __av_add_vdev(v, tar_vdev);
  if(dd->major == -1) {
    (*tar_vdev->destroy)(tar_vdev);
    return -1;
  }

  return dd->major;
}
