/*  
    AVFS: A Virtual File System Library
    Copyright (C) 1998  Miklos Szeredi (mszeredi@inf.bme.hu)
    
    This file can be distributed either under the GNU LGPL, or under
    the GNU GPL. See the file COPYING.LIB and COPYING. 

    ZFILT

    This module handles files compressed with the "zip" algorithm
*/

#include "zfilt.h"
#include "zutil.h"

#include <string.h>

#define INBUFSIZ 16384
#define OUTBUFSIZ 16384

typedef struct {
  struct zfilt_params fp;
  void *devinfo;

  z_stream stream;

  avoff_t rptr;

  avqbyte crc;
  avqbyte shoulbe_crc;

  char inbuf[INBUFSIZ];
  char outbuf[OUTBUFSIZ];

  int eof;
  int do_cache;

  int cachefd;
  avoff_t cacheptr;
} zfiltdi;


static void reset_zfilt_params(zfiltdi *di)
{
  di->stream.avail_in  = 0;
  di->rptr = 0;
  di->eof = 0;
  di->crc = crc32(0L, Z_NULL, 0);
}

static avssize_t read_from_zip(ave *v, zfiltdi *di, char *buf, avsize_t nbyte)
{
  int res;
  int streamend;
  avsize_t numread;

  di->stream.next_out = buf;
  di->stream.avail_out = nbyte;
  
  streamend = 0;
  while(di->stream.avail_out != 0) {
    
    if(di->stream.avail_in == 0 && !di->eof) {
      res = (*di->fp.readfunc)(v, di->devinfo, di->inbuf, INBUFSIZ);
      if(res == -1) return -1;
      if(res == 0) di->eof = 1;
      
      di->stream.next_in = di->inbuf;
      di->stream.avail_in = res;
    }

    res = inflate(&(di->stream), Z_NO_FLUSH);
    
    if(res == Z_STREAM_END) {
      streamend = 1;
      break;
    }
    if(res != Z_OK) {
      switch(res) {
      case Z_DATA_ERROR:
	__av_log(AVLOG_WARNING, "Zfilt: read: Corrupted data\n");
	v->errn = EIO;
	break;

      case Z_BUF_ERROR:
	__av_log(AVLOG_WARNING, "Zfilt: read: No progress; corrupted data\n");
	v->errn = EIO;
	break;

      case Z_MEM_ERROR:
	v->errn = ENOMEM;
	break;

      case Z_STREAM_ERROR:
	__av_log(AVLOG_WARNING, "Zfilt: read: internal error\n");
	v->errn = EFAULT;
	break;

      default:
	__av_log(AVLOG_WARNING, "Zfilt: read: unknown error\n");
	v->errn = EFAULT;
	break;

      }
      return -1;
    }
  }
  
  numread = nbyte - di->stream.avail_out;

  if(numread) di->crc = crc32(di->crc, buf, numread);

  if(streamend && di->crc != di->shoulbe_crc) {
    __av_log(AVLOG_WARNING, "Zfilt: read: CRC error\n");
    v->errn = EIO;
    return -1;
  }
  
  if(!numread) return 0;
  
  if(di->cachefd != -1) {
    if(di->cacheptr != di->rptr) 
      di->cacheptr = __av_locallseek(DUMMYV, di->cachefd, di->rptr, 
				     AVSEEK_SET);
    if(di->cacheptr != di->rptr || 
       __av_localwrite(DUMMYV, di->cachefd, buf, numread) != 
       (avssize_t) numread) {
      
      __av_log(AVLOG_WARNING, "Zfilt: Error writing to cachefile");
      __av_localclose(DUMMYV, di->cachefd);
      di->cachefd = -1;
    }
    else di->cacheptr += numread;
  }
  di->rptr += numread;

  return numread;
}


static avssize_t read_unneded(ave *v, zfiltdi *di, avsize_t nbyte)
{
  avssize_t res;
  avssize_t numread;
  
  numread = 0;
  while(nbyte > 0) {
    res = read_from_zip(v, di, di->outbuf, AV_MIN(nbyte, OUTBUFSIZ));
    if(res == -1) return -1;
    if(res == 0) break;

    numread += res;
    nbyte -= res;
  }

  return numread;
}

avssize_t __av_zfilt_read(ave *v, void *devinfo, char *buf, avoff_t offset,
			  avsize_t nbyte)
{
  zfiltdi *di = (zfiltdi *) devinfo;
  avssize_t rres, numread;

  numread = 0;

  if(offset < di->rptr && di->cachefd != -1) {
    avoff_t nact, avail;
    
    avail = di->rptr - offset;
    nact = AV_MIN((avoff_t) nbyte, avail);

    if(di->cacheptr != offset) 
      di->cacheptr = __av_locallseek(DUMMYV, di->cachefd, offset, 
				     AVSEEK_SET);
    
    if(di->cacheptr != offset || 
       __av_localread(DUMMYV, di->cachefd, buf, nact) != nact) {

      __av_log(AVLOG_WARNING, "Zfilt: Error reading from cachefile");
      __av_localclose(DUMMYV, di->cachefd);
      di->cachefd = -1;
    }
    else {
      di->cacheptr += nact;
      numread += nact;
      offset += nact;
      buf += nact;
      nbyte -= nact;
    }
  }

  if(nbyte == 0) return numread;

  if(offset < di->rptr) {
    if((*di->fp.lseekfunc)(v, di->devinfo, 0, AVSEEK_SET) == -1) 
      return -1;

    inflateReset(&(di->stream));
    
    reset_zfilt_params(di);

    /* If the cachefile is not opened the first time, it should be
       opened the second time */
    if(!di->do_cache) {
      di->do_cache = 1;
      AVDEB((AVLOG_DEBUG, "Zfilt: Opening cachefile (2)"));
      di->cachefd = __av_get_tmpfd(DUMMYV);
      if(di->cachefd == -1) 
	__av_log(AVLOG_WARNING, "Zfilt: Error opening cachefile");
    }
  }

  if(offset != di->rptr) {
    rres = read_unneded(v, di, offset - di->rptr);
    if(rres == -1) return -1;
    
    if(offset != di->rptr) return 0;
  }

  rres = read_from_zip(v, di, buf, nbyte);

  if(rres == -1) return -1;
  
  numread += rres;
  
  return numread;
}

void *__av_zfilt_open(ave *v, struct zfilt_params *fp, void *devinfo, 
		      avqbyte crc, int do_cache)
{
  zfiltdi *di;
  int res;

  AV_NEW(v, di);
  if(di == AVNULL) return AVNULL;

  di->fp = *fp;
  di->devinfo = devinfo;

  di->stream.zalloc    = NULL;
  di->stream.zfree     = NULL;
  di->stream.opaque    = NULL;
  di->stream.next_in   = NULL;
  di->stream.next_out  = NULL;
  di->stream.avail_in  = 0;
  di->stream.avail_out = 0;

  di->shoulbe_crc = crc;
  di->do_cache = do_cache;

  reset_zfilt_params(di);

  res = inflateInit2(&(di->stream), -MAX_WBITS);
  if(res != Z_OK) {
    if(res == Z_MEM_ERROR)
      v->errn = ENOMEM;
    else 
      v->errn = EFAULT;

    __av_log(AVLOG_WARNING, "Zfilt: inflateInit failed");
    __av_free(di);

    return AVNULL;
  }

  if(do_cache) {
    AVDEB((AVLOG_DEBUG, "Zfilt: Opening cachefile (2)"));
    di->cachefd = __av_get_tmpfd(DUMMYV);
    if(di->cachefd == -1) 
      __av_log(AVLOG_WARNING, "Zfilt: Error opening cachefile");
  }
  else 
    di->cachefd = -1;

  di->cacheptr = 0;

  return (void *) di;
}
                        
int __av_zfilt_close(ave *v, void *devinfo)
{
  int res = 0;
  zfiltdi *di = (zfiltdi *) devinfo;

  inflateEnd(&(di->stream));
  if(di->cachefd != -1) __av_localclose(DUMMYV, di->cachefd);
  __av_free(di);

  return res;
}
